# Copyright 2016-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotOperatorReverseDepsTestCase(TestCase):
    def testSlotOperatorReverseDeps(self):
        ebuilds = {
            "media-libs/mesa-11.2.2": {
                "EAPI": "6",
                "SLOT": "0",
                "RDEPEND": ">=sys-devel/llvm-3.6.0:=",
            },
            "sys-devel/clang-3.7.1-r100": {
                "EAPI": "6",
                "SLOT": "0/3.7",
                "RDEPEND": "~sys-devel/llvm-3.7.1",
            },
            "sys-devel/clang-3.8.0-r100": {
                "EAPI": "6",
                "SLOT": "0/3.8",
                "RDEPEND": "~sys-devel/llvm-3.8.0",
            },
            "sys-devel/llvm-3.7.1-r2": {
                "EAPI": "6",
                "SLOT": "0/3.7.1",
                "PDEPEND": "=sys-devel/clang-3.7.1-r100",
            },
            "sys-devel/llvm-3.8.0-r2": {
                "EAPI": "6",
                "SLOT": "0/3.8.0",
                "PDEPEND": "=sys-devel/clang-3.8.0-r100",
            },
        }

        installed = {
            "media-libs/mesa-11.2.2": {
                "EAPI": "6",
                "SLOT": "0",
                "RDEPEND": ">=sys-devel/llvm-3.6.0:0/3.7.1=",
            },
            "sys-devel/clang-3.7.1-r100": {
                "EAPI": "6",
                "SLOT": "0/3.7",
                "RDEPEND": "~sys-devel/llvm-3.7.1",
            },
            "sys-devel/llvm-3.7.1-r2": {
                "EAPI": "6",
                "SLOT": "0/3.7.1",
                "PDEPEND": "=sys-devel/clang-3.7.1-r100",
            },
        }

        world = ["media-libs/mesa"]

        test_cases = (
            # Test bug #584626, where an llvm update is missed due to
            # the check_reverse_dependencies function seeing that
            # updating llvm will break a dependency of the installed
            # version of clang (though a clang update is available).
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[
                    "sys-devel/llvm-3.8.0-r2",
                    "sys-devel/clang-3.8.0-r100",
                    "media-libs/mesa-11.2.2",
                ],
            ),
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--update": True,
                    "--deep": True,
                    "--ignore-built-slot-operator-deps": "y",
                },
                success=True,
                mergelist=[
                    "sys-devel/llvm-3.8.0-r2",
                    "sys-devel/clang-3.8.0-r100",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()


class SlotOperatorReverseDepsLibGit2TestCase(TestCase):
    def testSlotOperatorReverseDepsLibGit2(self):
        """
        Test bug #717140, where the depgraph _slot_operator_update_probe
        method ignored <dev-libs/libgit2-1:0= dependency and tried to
        trigger an upgrade to dev-libs/libgit2-1.0.0-r1, ultimately
        resulting in an undesirable downgrade to dev-libs/libgit2-0.28.4-r1.
        """

        ebuilds = {
            "dev-libs/libgit2-0.28.4-r1": {
                "EAPI": "7",
                "SLOT": "0/28",
            },
            "dev-libs/libgit2-0.99.0-r1": {
                "EAPI": "7",
                "SLOT": "0/0.99",
            },
            "dev-libs/libgit2-1.0.0-r1": {
                "EAPI": "7",
                "SLOT": "0/1.0",
            },
            "dev-libs/libgit2-glib-0.28.0.1": {
                "EAPI": "7",
                "SLOT": "0",
                "RDEPEND": "<dev-libs/libgit2-0.29:0= >=dev-libs/libgit2-0.26.0",
            },
            "dev-libs/libgit2-glib-0.99.0.1": {
                "EAPI": "7",
                "SLOT": "0",
                "RDEPEND": "<dev-libs/libgit2-1:0= >=dev-libs/libgit2-0.26.0",
            },
            "dev-vcs/gitg-3.32.1-r1": {
                "EAPI": "7",
                "SLOT": "0",
                "RDEPEND": "dev-libs/libgit2:= >=dev-libs/libgit2-glib-0.27 <dev-libs/libgit2-glib-1",
            },
        }

        installed = {
            "dev-libs/libgit2-0.99.0-r1": {
                "EAPI": "7",
                "SLOT": "0/0.99",
            },
            "dev-libs/libgit2-glib-0.99.0.1": {
                "EAPI": "7",
                "SLOT": "0",
                "RDEPEND": "<dev-libs/libgit2-1:0/0.99= >=dev-libs/libgit2-0.26.0",
            },
            "dev-vcs/gitg-3.32.1-r1": {
                "EAPI": "7",
                "SLOT": "0",
                "RDEPEND": "dev-libs/libgit2:0/0.99= >=dev-libs/libgit2-glib-0.27 <dev-libs/libgit2-glib-1",
            },
        }

        world = ["dev-vcs/gitg"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                # mergelist = ['dev-libs/libgit2-0.28.4-r1', 'dev-libs/libgit2-glib-0.99.0.1', 'dev-vcs/gitg-3.32.1-r1'],
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()


class SlotOperatorReverseDepsVirtualTestCase(TestCase):
    def testSlotOperatorReverseDepsVirtual(self):
        """
        Demonstrate bug #764764, where slot operator rebuilds were
        not triggered for reverse deps of virtual/dist-kernel.
        """

        ebuilds = {
            "app-emulation/virtualbox-modules-6.1.16-r1": {
                "EAPI": "7",
                "DEPEND": "virtual/dist-kernel",
                "RDEPEND": "virtual/dist-kernel:=",
            },
            "sys-kernel/gentoo-kernel-5.10.6": {
                "EAPI": "7",
                "SLOT": "5.10.6",
            },
            "sys-kernel/gentoo-kernel-5.10.5": {
                "EAPI": "7",
                "SLOT": "5.10.5",
            },
            "virtual/dist-kernel-5.10.5": {
                "EAPI": "7",
                "SLOT": "0/5.10.5",
                "RDEPEND": "~sys-kernel/gentoo-kernel-5.10.5",
            },
            "virtual/dist-kernel-5.10.6": {
                "EAPI": "7",
                "SLOT": "0/5.10.6",
                "RDEPEND": "~sys-kernel/gentoo-kernel-5.10.6",
            },
            "x11-drivers/nvidia-drivers-460.32.03": {
                "EAPI": "7",
                "DEPEND": "virtual/dist-kernel",
                "RDEPEND": "virtual/dist-kernel:=",
            },
        }

        installed = {
            "app-emulation/virtualbox-modules-6.1.16-r1": {
                "EAPI": "7",
                "DEPEND": "virtual/dist-kernel",
                "RDEPEND": "virtual/dist-kernel:0/5.10.5=",
            },
            "sys-kernel/gentoo-kernel-5.10.5": {
                "EAPI": "7",
                "SLOT": "5.10.5",
            },
            "virtual/dist-kernel-5.10.5": {
                "EAPI": "7",
                "SLOT": "0/5.10.5",
                "RDEPEND": "~sys-kernel/gentoo-kernel-5.10.5",
            },
            "x11-drivers/nvidia-drivers-460.32.03": {
                "EAPI": "7",
                "DEPEND": "virtual/dist-kernel",
                "RDEPEND": "virtual/dist-kernel:0/5.10.5=",
            },
        }

        world = ["app-emulation/virtualbox-modules", "x11-drivers/nvidia-drivers"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@world"],
                options={"--update": True, "--deep": True},
                success=True,
                mergelist=[
                    "sys-kernel/gentoo-kernel-5.10.6",
                    "virtual/dist-kernel-5.10.6",
                    "app-emulation/virtualbox-modules-6.1.16-r1",
                    "x11-drivers/nvidia-drivers-460.32.03",
                ],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2013-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2


import portage
from portage import os, _encodings
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)

from portage.glsa import GlsaFormatException


class SecuritySetTestCase(TestCase):
    glsa_template = """\
<?xml version="1.0" encoding="UTF-8"?>
<?xml-stylesheet href="/xsl/glsa.xsl" type="text/xsl"?>
<?xml-stylesheet href="/xsl/guide.xsl" type="text/xsl"?>
<!DOCTYPE glsa SYSTEM "http://www.gentoo.org/dtd/glsa.dtd">
<glsa id="%(glsa_id)s">
  <title>%(pkgname)s: Multiple vulnerabilities</title>
  <synopsis>Multiple vulnerabilities have been found in %(pkgname)s.
  </synopsis>
  <product type="ebuild">%(pkgname)s</product>
  <announced>January 18, 2013</announced>
  <revised count="1">January 18, 2013</revised>
  <bug>55555</bug>
  <access>remote</access>
  <affected>
    <package name="%(cp)s" auto="yes" arch="%(arch)s">
      <unaffected range="%(unaffected_range)s">%(unaffected_version)s</unaffected>
      <vulnerable range="%(affected_range)s">%(affected_version)s</vulnerable>
    </package>
  </affected>
  <background>
    <p>%(pkgname)s is software package.</p>
  </background>
  <description>
    <p>Multiple vulnerabilities have been discovered in %(pkgname)s.
    </p>
  </description>
  <impact type="normal">
    <p>A remote attacker could exploit these vulnerabilities.</p>
  </impact>
  <workaround>
    <p>There is no known workaround at this time.</p>
  </workaround>
  <resolution>
    <p>All %(pkgname)s users should upgrade to the latest version:</p>
    <code>
      # emerge --sync
      # emerge --ask --oneshot --verbose "&gt;=%(cp)s-%(unaffected_version)s"
    </code>
  </resolution>
  <references>
  </references>
</glsa>
"""

    def _must_skip(self):
        try:
            __import__("xml.etree.ElementTree")
            __import__("xml.parsers.expat").parsers.expat.ExpatError
        except (AttributeError, ImportError):
            self.skipTest("python is missing xml support")

    def write_glsa_test_case(self, glsa_dir, glsa):
        with open(
            os.path.join(glsa_dir, "glsa-" + glsa["glsa_id"] + ".xml"),
            encoding=_encodings["repo.content"],
            mode="w",
        ) as f:
            f.write(self.glsa_template % glsa)

    def testSecuritySet(self):
        self._must_skip()

        ebuilds = {
            "cat/A-vulnerable-2.2": {"KEYWORDS": "x86"},
            "cat/B-not-vulnerable-4.5": {"KEYWORDS": "x86"},
        }

        installed = {
            "cat/A-vulnerable-2.1": {"KEYWORDS": "x86"},
            "cat/B-not-vulnerable-4.4": {"KEYWORDS": "x86"},
        }

        glsas = (
            {
                "glsa_id": "201301-01",
                "pkgname": "A-vulnerable",
                "cp": "cat/A-vulnerable",
                "unaffected_range": "ge",
                "affected_range": "lt",
                "unaffected_version": "2.2",
                "affected_version": "2.2",
                "arch": "*",
            },
            {
                "glsa_id": "201301-02",
                "pkgname": "B-not-vulnerable",
                "cp": "cat/B-not-vulnerable",
                "unaffected_range": "ge",
                "affected_range": "lt",
                "unaffected_version": "4.4",
                "affected_version": "4.4",
                "arch": "*",
            },
            {
                "glsa_id": "201301-03",
                "pkgname": "NotInstalled",
                "cp": "cat/NotInstalled",
                "unaffected_range": "ge",
                "affected_range": "lt",
                "unaffected_version": "3.5",
                "affected_version": "3.5",
                "arch": "*",
            },
        )

        world = ["cat/A"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@security"],
                options={},
                success=True,
                mergelist=["cat/A-vulnerable-2.2"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )

        try:
            portdb = playground.trees[playground.eroot]["porttree"].dbapi
            glsa_dir = os.path.join(
                portdb.repositories["test_repo"].location, "metadata", "glsa"
            )
            portage.util.ensure_dirs(glsa_dir)
            for glsa in glsas:
                self.write_glsa_test_case(glsa_dir, glsa)

            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testStatelessSecuritySet(self):
        # Tests which don't rely on the GLSA being fixed. This allows
        # testing the format parsing with a bit more flexibility (no
        # need to keep inventing packages).

        self._must_skip()

        ebuilds = {
            "cat/A-vulnerable-2.2": {"KEYWORDS": "x86"},
            "cat/B-not-vulnerable-4.5": {"KEYWORDS": "x86"},
        }

        installed = {
            "cat/A-vulnerable-2.1": {"KEYWORDS": "x86"},
            "cat/B-not-vulnerable-4.4": {"KEYWORDS": "x86"},
        }

        glsas = (
            {
                "glsa_id": "201301-04",
                "pkgname": "A-vulnerable",
                "cp": "cat/A-vulnerable",
                "unaffected_range": "ge",
                "affected_range": "lt",
                "unaffected_version": "2.2",
                "affected_version": "2.2",
                # Use an invalid delimiter (comma)
                "arch": "amd64,sparc",
            },
            {
                "glsa_id": "201301-05",
                "pkgname": "A-vulnerable",
                "cp": "cat/A-vulnerable",
                "unaffected_range": "ge",
                "affected_range": "lt",
                "unaffected_version": "2.2",
                "affected_version": "2.2",
                # Use an invalid arch (~arch)
                "arch": "~amd64",
            },
            {
                "glsa_id": "201301-06",
                "pkgname": "A-vulnerable",
                "cp": "cat/A-vulnerable",
                "unaffected_range": "ge",
                "affected_range": "lt",
                "unaffected_version": "2.2",
                "affected_version": "2.2",
                # Two valid arches followed by an invalid one
                "arch": "amd64 sparc $$$$",
            },
            {
                "glsa_id": "201301-07",
                "pkgname": "A-vulnerable",
                "cp": "cat/A-vulnerable",
                "unaffected_range": "None",
                "affected_range": "lt",
                "unaffected_version": "2.2",
                "affected_version": "2.2",
                "arch": "*",
            },
        )

        world = ["cat/A"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["@security"],
                success=True,
                mergelist=["cat/A-vulnerable-2.2"],
            ),
        )

        # Give each GLSA a clean slate
        for glsa in glsas:
            playground = ResolverPlayground(
                ebuilds=ebuilds, installed=installed, world=world, debug=False
            )

            try:
                portdb = playground.trees[playground.eroot]["porttree"].dbapi
                glsa_dir = os.path.join(
                    portdb.repositories["test_repo"].location, "metadata", "glsa"
                )
                portage.util.ensure_dirs(glsa_dir)

                self.write_glsa_test_case(glsa_dir, glsa)

                with self.assertRaises(GlsaFormatException):
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
            finally:
                playground.cleanup()



# elog/mod_custom.py - elog dispatch module
# Copyright 2006-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import types

import portage
import portage.elog.mod_save
import portage.exception
import portage.process
from portage.util.futures import asyncio

# Since elog_process is typically called while the event loop is
# running, hold references to spawned processes and wait for them
# asynchronously, ultimately waiting for them if necessary when
# the AsyncioEventLoop _close_main method calls _async_finalize
# via portage.process.run_coroutine_exitfuncs().
_proc_refs = None


def _get_procs() -> list[tuple[portage.process.MultiprocessingProcess, asyncio.Future]]:
    """
    Return list of (proc, asyncio.ensure_future(proc.wait())) which is not
    inherited from the parent after fork.
    """
    global _proc_refs
    if _proc_refs is None or _proc_refs.pid != portage.getpid():
        _proc_refs = types.SimpleNamespace(pid=portage.getpid(), procs=[])
        portage.process.atexit_register(_async_finalize)
    return _proc_refs.procs


def process(mysettings, key, logentries, fulltext):
    elogfilename = portage.elog.mod_save.process(mysettings, key, logentries, fulltext)

    if not mysettings.get("PORTAGE_ELOG_COMMAND"):
        raise portage.exception.MissingParameter(
            "!!! Custom logging requested but PORTAGE_ELOG_COMMAND is not defined"
        )
    else:
        mylogcmd = mysettings["PORTAGE_ELOG_COMMAND"]
        mylogcmd = mylogcmd.replace("${LOGFILE}", elogfilename)
        mylogcmd = mylogcmd.replace("${PACKAGE}", key)
        loop = asyncio.get_event_loop()
        proc = portage.process.spawn_bash(mylogcmd, returnproc=True)
        procs = _get_procs()
        procs.append((proc, asyncio.ensure_future(proc.wait(), loop=loop)))
        for index, (proc, waiter) in reversed(list(enumerate(procs))):
            if not waiter.done():
                continue
            del procs[index]
            if waiter.result() != 0:
                raise portage.exception.PortageException(
                    f"!!! PORTAGE_ELOG_COMMAND failed with exitcode {waiter.result()}"
                )


async def _async_finalize():
    """
    Async finalize is preferred, since we can wait for process exit status.
    """
    procs = _get_procs()
    while procs:
        proc, waiter = procs.pop()
        if (await waiter) != 0:
            raise portage.exception.PortageException(
                f"!!! PORTAGE_ELOG_COMMAND failed with exitcode {waiter.result()}"
            )


def finalize():
    """
    NOTE: This raises PortageException if there are any processes
    still running, so it's better to use _async_finalize instead
    (invoked via portage.process.run_coroutine_exitfuncs() in
    the AsyncioEventLoop _close_main method).
    """
    procs = _get_procs()
    while procs:
        proc, waiter = procs.pop()
        if not waiter.done():
            waiter.cancel()
            proc.terminate()
            raise portage.exception.PortageException(
                f"!!! PORTAGE_ELOG_COMMAND was killed after it was found running in the background (pid {proc.pid})"
            )
        elif waiter.result() != 0:
            raise portage.exception.PortageException(
                f"!!! PORTAGE_ELOG_COMMAND failed with exitcode {waiter.result()}"
            )



# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.dep import Atom
from portage.dep.libc import strip_libc_deps
from portage.tests import TestCase


class LibcUtilStripDeps(TestCase):
    def testStripSimpleDeps(self):
        """
        Test that we strip a basic libc dependency out and return
        a list of dependencies without it in there.
        """

        libc_dep = [Atom("=sys-libs/glibc-2.38")]

        original_deps = (
            [
                Atom("=sys-libs/glibc-2.38"),
                Atom("=app-misc/foo-1.2.3"),
            ],
            [
                Atom("=sys-libs/glibc-2.38"),
            ],
            [
                Atom("=app-misc/foo-1.2.3"),
                Atom("=app-misc/bar-1.2.3"),
            ],
        )

        for deplist in original_deps:
            strip_libc_deps(deplist, libc_dep)

            self.assertFalse(
                all(libc in deplist for libc in libc_dep),
                "Stripped deplist contains a libc candidate",
            )

    def testStripComplexRealizedDeps(self):
        """
        Test that we strip pathological libc dependencies out and return
        a list of dependencies without it in there.
        """

        # This shouldn't really happen for a 'realized' dependency, but
        # we shouldn't crash if it happens anyway.
        libc_dep = [Atom("=sys-libs/glibc-2.38*[p]")]

        original_deps = (
            [
                Atom("=sys-libs/glibc-2.38[x]"),
                Atom("=app-misc/foo-1.2.3"),
            ],
            [
                Atom("=sys-libs/glibc-2.38[p]"),
            ],
            [
                Atom("=app-misc/foo-1.2.3"),
                Atom("=app-misc/bar-1.2.3"),
            ],
        )

        for deplist in original_deps:
            strip_libc_deps(deplist, libc_dep)

            self.assertFalse(
                all(libc in deplist for libc in libc_dep),
                "Stripped deplist contains a libc candidate",
            )

    def testStripNonRealizedDeps(self):
        """
        Check that we strip non-realized libc deps.
        """

        libc_dep = [Atom("sys-libs/glibc:2.2=")]
        original_deps = [Atom(">=sys-libs/glibc-2.38-r7")]

        strip_libc_deps(original_deps, libc_dep)
        self.assertFalse(original_deps, "(g)libc dep was not stripped")



# Copyright 1999-2013 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import platform
import sys
import time

from portage.output import darkgreen, green


class stdout_spinner:
    scroll_msgs = [
        "Gentoo Rocks (" + platform.system() + ")",
        "Thank you for using Gentoo. :)",
        "Are you actually trying to read this?",
        "How many times have you stared at this?",
        "We are generating the cache right now",
        "You are paying too much attention.",
        "A theory is better than its explanation.",
        "Phasers locked on target, Captain.",
        "Thrashing is just virtual crashing.",
        "To be is to program.",
        "Real Users hate Real Programmers.",
        "When all else fails, read the instructions.",
        "Functionality breeds Contempt.",
        "The future lies ahead.",
        "3.1415926535897932384626433832795028841971694",
        "Sometimes insanity is the only alternative.",
        "Inaccuracy saves a world of explanation.",
    ]

    twirl_sequence = "/-\\|/-\\|/-\\|/-\\|\\-/|\\-/|\\-/|\\-/|"

    def __init__(self):
        self.spinpos = 0
        self.update = self.update_twirl
        self.scroll_sequence = self.scroll_msgs[
            int(time.time() * 100) % len(self.scroll_msgs)
        ]
        self.last_update = 0
        self.min_display_latency = 0.05
        self.start_time = None

    def _return_early(self):
        """
        Flushing ouput to the tty too frequently wastes cpu time. Therefore,
        each update* method should return without doing any output when this
        method returns True.
        """
        cur_time = time.time()
        if cur_time - self.last_update < self.min_display_latency:
            return True
        self.last_update = cur_time
        return False

    def update_basic(self):
        self.spinpos = (self.spinpos + 1) % 500
        if self._return_early():
            return True
        if (self.spinpos % 100) == 0:
            if self.spinpos == 0:
                sys.stdout.write(". ")
            else:
                sys.stdout.write(".")
        sys.stdout.flush()
        return True

    def update_scroll(self):
        if self._return_early():
            return True
        if self.spinpos >= len(self.scroll_sequence):
            sys.stdout.write(
                darkgreen(
                    " \b\b\b"
                    + self.scroll_sequence[
                        len(self.scroll_sequence)
                        - 1
                        - (self.spinpos % len(self.scroll_sequence))
                    ]
                )
            )
        else:
            sys.stdout.write(green("\b " + self.scroll_sequence[self.spinpos]))
        sys.stdout.flush()
        self.spinpos = (self.spinpos + 1) % (2 * len(self.scroll_sequence))
        return True

    def update_twirl(self):
        self.spinpos = (self.spinpos + 1) % len(self.twirl_sequence)
        if self._return_early():
            return True
        sys.stdout.write("\b\b " + self.twirl_sequence[self.spinpos])
        sys.stdout.flush()
        return True

    def update_quiet(self):
        return True



# Copyright 2023 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import argparse
import shlex
from typing import Optional, Callable  # ,  Self

from portage.const import (
    SUPPORTED_GENTOO_BINPKG_FORMATS,
    BASH_BINARY,
    BINREPOS_CONF_FILE,
)
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.cache.mappings import Mapping
from portage.tests.util.test_socks5 import AsyncHTTPServer
from portage import os
from portage import shutil
from portage.util.futures import asyncio
from portage.tests import cnf_bindir, cnf_sbindir
from portage.process import find_binary
from portage.util import find_updated_config_files
import portage

import pytest


_INSTALL_SOMETHING = """
S="${WORKDIR}"

pkg_pretend() {
	einfo "called pkg_pretend for $CATEGORY/$PF"
}

src_install() {
	einfo "installing something..."
	insinto /usr/lib/${P}
	echo "blah blah blah" > "${T}"/regular-file
	doins "${T}"/regular-file
	dosym regular-file /usr/lib/${P}/symlink || die

	# Test CONFIG_PROTECT
	insinto /etc
	newins "${T}"/regular-file ${PN}-${SLOT%/*}

	# Test code for bug #381629, using a copyright symbol encoded with latin-1.
	# We use $(printf "\\xa9") rather than $'\\xa9', since printf apparently
	# works in any case, while $'\\xa9' transforms to \\xef\\xbf\\xbd under
	# some conditions. TODO: Find out why it transforms to \\xef\\xbf\\xbd when
	# running tests for Python 3.2 (even though it's bash that is ultimately
	# responsible for performing the transformation).
	local latin_1_dir=/usr/lib/${P}/latin-1-$(printf "\\xa9")-directory
	insinto "${latin_1_dir}"
	echo "blah blah blah" > "${T}"/latin-1-$(printf "\\xa9")-regular-file || die
	doins "${T}"/latin-1-$(printf "\\xa9")-regular-file
	dosym latin-1-$(printf "\\xa9")-regular-file ${latin_1_dir}/latin-1-$(printf "\\xa9")-symlink || die

	call_has_and_best_version
}

pkg_config() {
	einfo "called pkg_config for $CATEGORY/$PF"
}

pkg_info() {
	einfo "called pkg_info for $CATEGORY/$PF"
}

pkg_preinst() {
	if ! ___eapi_best_version_and_has_version_support_-b_-d_-r; then
		# The BROOT variable is unset during pkg_* phases for EAPI 7,
		# therefore best/has_version -b is expected to fail if we attempt
		# to call it for EAPI 7 here.
		call_has_and_best_version
	fi
}

call_has_and_best_version() {
	local root_arg
	if ___eapi_best_version_and_has_version_support_-b_-d_-r; then
		root_arg="-b"
	else
		root_arg="--host-root"
	fi
	einfo "called ${EBUILD_PHASE_FUNC} for $CATEGORY/$PF"
	einfo "EPREFIX=${EPREFIX}"
	einfo "PORTAGE_OVERRIDE_EPREFIX=${PORTAGE_OVERRIDE_EPREFIX}"
	einfo "ROOT=${ROOT}"
	einfo "EROOT=${EROOT}"
	einfo "SYSROOT=${SYSROOT}"
	einfo "ESYSROOT=${ESYSROOT}"
	einfo "BROOT=${BROOT}"
	# Test that has_version and best_version work correctly with
	# prefix (involves internal ROOT -> EROOT calculation in order
	# to support ROOT override via the environment with EAPIs 3
	# and later which support prefix).
	if has_version $CATEGORY/$PN:$SLOT ; then
		einfo "has_version detects an installed instance of $CATEGORY/$PN:$SLOT"
		einfo "best_version reports that the installed instance is $(best_version $CATEGORY/$PN:$SLOT)"
	else
		einfo "has_version does not detect an installed instance of $CATEGORY/$PN:$SLOT"
	fi
	if [[ ${EPREFIX} != ${PORTAGE_OVERRIDE_EPREFIX} ]] ; then
		if has_version ${root_arg} $CATEGORY/$PN:$SLOT ; then
			einfo "has_version ${root_arg} detects an installed instance of $CATEGORY/$PN:$SLOT"
			einfo "best_version ${root_arg} reports that the installed instance is $(best_version ${root_arg} $CATEGORY/$PN:$SLOT)"
		else
			einfo "has_version ${root_arg} does not detect an installed instance of $CATEGORY/$PN:$SLOT"
		fi
	fi
}

"""

_AVAILABLE_EBUILDS = {
    "dev-libs/A-1": {
        "EAPI": "5",
        "IUSE": "+flag",
        "KEYWORDS": "x86",
        "LICENSE": "GPL-2",
        "MISC_CONTENT": _INSTALL_SOMETHING,
        "RDEPEND": "flag? ( dev-libs/B[flag] )",
    },
    "dev-libs/B-1": {
        "EAPI": "5",
        "IUSE": "+flag",
        "KEYWORDS": "x86",
        "LICENSE": "GPL-2",
        "MISC_CONTENT": _INSTALL_SOMETHING,
    },
    "dev-libs/C-1": {
        "EAPI": "7",
        "KEYWORDS": "~x86",
        "RDEPEND": "dev-libs/D[flag]",
        "MISC_CONTENT": _INSTALL_SOMETHING,
    },
    "dev-libs/D-1": {
        "EAPI": "7",
        "KEYWORDS": "~x86",
        "IUSE": "flag",
        "MISC_CONTENT": _INSTALL_SOMETHING,
    },
    "virtual/foo-0": {
        "EAPI": "5",
        "KEYWORDS": "x86",
        "LICENSE": "GPL-2",
    },
}

_INSTALLED_EBUILDS = {
    "dev-libs/A-1": {
        "EAPI": "5",
        "IUSE": "+flag",
        "KEYWORDS": "x86",
        "LICENSE": "GPL-2",
        "RDEPEND": "flag? ( dev-libs/B[flag] )",
        "USE": "flag",
    },
    "dev-libs/B-1": {
        "EAPI": "5",
        "IUSE": "+flag",
        "KEYWORDS": "x86",
        "LICENSE": "GPL-2",
        "USE": "flag",
    },
    "dev-libs/depclean-me-1": {
        "EAPI": "5",
        "IUSE": "",
        "KEYWORDS": "x86",
        "LICENSE": "GPL-2",
        "USE": "",
    },
    "app-misc/depclean-me-1": {
        "EAPI": "5",
        "IUSE": "",
        "KEYWORDS": "x86",
        "LICENSE": "GPL-2",
        "RDEPEND": "dev-libs/depclean-me",
        "USE": "",
    },
}


_BASELINE_COMMAND_SEQUENCE = [
    "emerge -1 dev-libs/A -v dev-libs/B",
    "emerge with quickpkg direct",
    "env-update",
    "portageq envvar",
    "etc-update",
    "dispatch-conf",
    "emerge --version",
    "emerge --info",
    "emerge --info --verbose",
    "emerge --list-sets",
    "emerge --check-news",
    "emerge --regen/--metadata",
    "misc package operations",
    "binhost emerge",
]

PORTAGE_PYTHON = portage._python_interpreter
NOOP = lambda: ...


class PortageCommand:
    """A class that represents a baseline test case command,
    including handling of environment and one-use arguments.
    """

    command = None
    name = None

    def __init__(
        self,
        *args: tuple[str],
        env_mod: Optional[dict[str, str]] = None,
        preparation: Optional[Callable[[], None]] = None,
        post_command: Optional[Callable[[], None]] = None,
    ) -> None:
        self.args = args
        self.env_mod = env_mod
        self.preparation = preparation
        self.post_command = post_command

    def __iter__(self):
        """To be able to call a function with ``*command`` as argument."""
        yield self

    @property
    def env(self) -> dict[str, str]:
        """This property returns the environment intended to be used
        with the current test command, including possible modifications.
        """
        try:
            base_environment = self.base_environment
        except AttributeError:
            base_environment = {}
        else:
            base_environment = base_environment.copy()
        if self.env_mod:
            base_environment.update(self.env_mod)
        return base_environment

    def __call__(self):  #  -> Self:
        if self.preparation:
            self.preparation()
        try:
            tuple_command = self.command + self.args
        except TypeError:
            # In case self.command is a string:
            tuple_command = (self.command,) + self.args
        return tuple_command

    def __bool__(self) -> bool:
        return bool(self.command)

    def check_command_result(self) -> None:
        if self.post_command:
            self.post_command()


class PortageCommandSequence:
    def __init__(self, *commands):
        self.commands = commands

    def __iter__(self):
        yield from self.commands


class Emerge(PortageCommand):
    name = "emerge"
    command = (PORTAGE_PYTHON, "-b", "-Wd", os.path.join(str(cnf_bindir), name))


class Noop(PortageCommand):
    name = "No-op"


class EnvUpdate(PortageCommand):
    name = "env-update"
    command = (PORTAGE_PYTHON, "-b", "-Wd", os.path.join(str(cnf_sbindir), name))


class DispatchConf(PortageCommand):
    name = "dispatch-conf"
    command = (
        PORTAGE_PYTHON,
        "-b",
        "-Wd",
        os.path.join(str(cnf_sbindir), name),
    )


class Ebuild(PortageCommand):
    name = "ebuild"
    command = (PORTAGE_PYTHON, "-b", "-Wd", os.path.join(str(cnf_bindir), name))


class Egencache(PortageCommand):
    name = "egencache"
    command = (
        PORTAGE_PYTHON,
        "-b",
        "-Wd",
        os.path.join(str(cnf_bindir), name),
    )


class Emaint(PortageCommand):
    name = "emaint"
    command = (PORTAGE_PYTHON, "-b", "-Wd", os.path.join(str(cnf_sbindir), name))


class EtcUpdate(PortageCommand):
    name = "etc-update"
    command = (BASH_BINARY, os.path.join(str(cnf_sbindir), name))


class Fixpackages(PortageCommand):
    name = "fixpackages"
    command = (
        PORTAGE_PYTHON,
        "-b",
        "-Wd",
        os.path.join(str(cnf_sbindir), name),
    )


class Portageq(PortageCommand):
    name = "portageq"
    command = (
        PORTAGE_PYTHON,
        "-b",
        "-Wd",
        os.path.join(str(cnf_bindir), name),
    )


class Quickpkg(PortageCommand):
    name = "quickpkg"
    command = (
        PORTAGE_PYTHON,
        "-b",
        "-Wd",
        os.path.join(str(cnf_bindir), name),
    )


class Regenworld(PortageCommand):
    name = "regenworld"
    command = (
        PORTAGE_PYTHON,
        "-b",
        "-Wd",
        os.path.join(str(cnf_sbindir), name),
    )


def pytest_generate_tests(metafunc):
    if "baseline_command" in metafunc.fixturenames:
        metafunc.parametrize(
            "baseline_command", _BASELINE_COMMAND_SEQUENCE, indirect=True
        )


def _have_python_xml():
    try:
        __import__("xml.etree.ElementTree")
        __import__("xml.parsers.expat").parsers.expat.ExpatError
    except (AttributeError, ImportError):
        return False
    return True


def _check_foo_file(pkgdir, filename, must_exist) -> None:
    assert (
        os.path.exists(os.path.join(pkgdir, "virtual", "foo", filename)) == must_exist
    )


def _check_number_of_protected_files(must_have, eroot, config_protect) -> None:
    assert must_have == len(
        list(find_updated_config_files(eroot, shlex.split(config_protect)))
    )


class BinhostContentMap(Mapping):
    def __init__(self, remote_path, local_path):
        self._remote_path = remote_path
        self._local_path = local_path

    def __getitem__(self, request_path):
        safe_path = os.path.normpath(request_path)
        if not safe_path.startswith(self._remote_path + "/"):
            raise KeyError(request_path)
        local_path = os.path.join(
            self._local_path, safe_path[len(self._remote_path) + 1 :]
        )
        try:
            with open(local_path, "rb") as f:
                return f.read()
        except OSError:
            raise KeyError(request_path)


@pytest.fixture(scope="module")
def async_loop():
    yield asyncio._wrap_loop()


@pytest.fixture(params=SUPPORTED_GENTOO_BINPKG_FORMATS, scope="function")
def playground(request, tmp_path_factory):
    """Fixture that provides instances of ``ResolverPlayground``
    each one with one supported value for ``BINPKG_FORMAT``."""
    binpkg_format = request.param
    playground = ResolverPlayground(
        ebuilds=_AVAILABLE_EBUILDS,
        installed=_INSTALLED_EBUILDS,
        debug=False,
        user_config={
            "make.conf": (f'BINPKG_FORMAT="{binpkg_format}"',),
        },
        eprefix=str(tmp_path_factory.mktemp("eprefix", numbered=True)),
    )
    yield playground
    playground.cleanup()


@pytest.fixture()
def binhost(playground, async_loop):
    settings = playground.settings
    eprefix = settings["EPREFIX"]
    binhost_dir = os.path.join(eprefix, "binhost")
    binhost_address = "127.0.0.1"
    binhost_remote_path = "/binhost"
    binhost_server = AsyncHTTPServer(
        binhost_address, BinhostContentMap(binhost_remote_path, binhost_dir), async_loop
    ).__enter__()
    binhost_uri = "http://{address}:{port}{path}".format(
        address=binhost_address,
        port=binhost_server.server_port,
        path=binhost_remote_path,
    )
    yield {"server": binhost_server, "uri": binhost_uri, "dir": binhost_dir}
    binhost_server.__exit__(None, None, None)


@pytest.fixture()
def _generate_all_baseline_commands(playground, binhost):
    """This fixture generates all the commands that
    ``test_portage_baseline`` will use.

    But, don't use this fixture directly, instead, use the
    ``baseline_command`` fixture. That improves performance a bit due to
    pytest caching (?).

    .. note::

       To add a new command, define it in the local ``test_commands``
       dict, if not yet defined, and add its key at the correct position
       in the ``_BASELINE_COMMAND_SEQUENCE`` list.
    """
    settings = playground.settings
    eprefix = settings["EPREFIX"]
    eroot = settings["EROOT"]
    trees = playground.trees
    pkgdir = playground.pkgdir
    portdb = trees[eroot]["porttree"].dbapi
    test_repo_location = settings.repositories["test_repo"].location
    var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
    cachedir = os.path.join(var_cache_edb, "dep")
    cachedir_pregen = os.path.join(test_repo_location, "metadata", "md5-cache")

    rm_binary = find_binary("rm")
    assert rm_binary is not None, "rm command not found"
    rm_cmd = (rm_binary,)

    egencache_extra_args = []
    if _have_python_xml():
        egencache_extra_args.append("--update-use-local-desc")

    test_ebuild = portdb.findname("dev-libs/A-1")
    assert test_ebuild is not None

    cross_prefix = os.path.join(eprefix, "cross_prefix")
    cross_root = os.path.join(eprefix, "cross_root")
    cross_eroot = os.path.join(cross_root, eprefix.lstrip(os.sep))

    binpkg_format = settings.get("BINPKG_FORMAT", SUPPORTED_GENTOO_BINPKG_FORMATS[0])
    assert binpkg_format in ("xpak", "gpkg")
    if binpkg_format == "xpak":
        foo_filename = "foo-0-1.xpak"
    elif binpkg_format == "gpkg":
        foo_filename = "foo-0-1.gpkg.tar"

    test_commands = {}

    if hasattr(argparse.ArgumentParser, "parse_intermixed_args"):
        parse_intermixed_command = Emerge(
            "--oneshot",
            "dev-libs/A",
            "-v",
            "dev-libs/A",
        )
    else:
        parse_intermixed_command = Noop()
    test_commands["emerge -1 dev-libs/A -v dev-libs/B"] = parse_intermixed_command

    quickpkg_direct_seq = [
        Emerge(
            "--usepkgonly",
            "--root",
            cross_root,
            "--quickpkg-direct=y",
            "--quickpkg-direct-root",
            "/",
            "dev-libs/A",
        ),
        # v needs ^
        Emerge(
            "--usepkgonly",
            "--quickpkg-direct=y",
            "--quickpkg-direct-root",
            cross_root,
            "dev-libs/A",
        ),
    ]
    test_commands["emerge with quickpkg direct"] = PortageCommandSequence(
        *quickpkg_direct_seq
    )

    test_commands["env-update"] = EnvUpdate()
    test_commands["portageq envvar"] = Portageq(
        "envvar",
        "-v",
        "CONFIG_PROTECT",
        "EROOT",
        "PORTAGE_CONFIGROOT",
        "PORTAGE_TMPDIR",
        "CHOST*",
        "USERLAND",
    )
    test_commands["etc-update"] = EtcUpdate()
    test_commands["dispatch-conf"] = DispatchConf()
    test_commands["emerge --version"] = Emerge("--version")
    test_commands["emerge --info"] = Emerge("--info")
    test_commands["emerge --info --verbose"] = Emerge("--info", "--verbose")
    test_commands["emerge --list-sets"] = Emerge("--list-sets")
    test_commands["emerge --check-news"] = Emerge("--check-news")

    def _rm_cachedir():
        shutil.rmtree(cachedir)

    def _rm_cachedir_and_pregen():
        _rm_cachedir()
        shutil.rmtree(cachedir_pregen)

    regen_seq = [
        Emerge("--regen", preparation=_rm_cachedir_and_pregen),
        Emerge(
            "--regen",
            env_mod={"FEATURES": "metadata-transfer"},
            preparation=_rm_cachedir,
        ),
        Egencache(
            "--repo",
            "test_repo",
            "--repositories-configuration",
            playground.settings.repositories.config_string(),
            "--update",
            *egencache_extra_args,
            preparation=_rm_cachedir,
        ),
        Emerge("--metadata", env_mod={"FEATURES": "metadata-transfer"}),
        Emerge(
            "--metadata",
            env_mod={"FEATURES": "metadata-transfer"},
            preparation=_rm_cachedir,
        ),
        Emerge("--metadata"),
        Emerge("--oneshot", "virtual/foo", preparation=_rm_cachedir),
        Emerge(
            "--unmerge",
            "virtual/foo",
            env_mod={"FEATURES": "unmerge-backup"},
            preparation=lambda: _check_foo_file(pkgdir, foo_filename, must_exist=False),
        ),
        Emerge(
            "--pretend",
            "dev-libs/A",
            preparation=lambda: _check_foo_file(pkgdir, foo_filename, must_exist=True),
        ),
    ]
    test_commands["emerge --regen/--metadata"] = PortageCommandSequence(*regen_seq)

    abcd_seq = [
        Ebuild(
            test_ebuild,
            "manifest",
            "clean",
            "package",
            "merge",
        ),
        Emerge(
            "--pretend",
            "--tree",
            "--complete-graph",
            "dev-libs/A",
        ),
        Emerge("-p", "dev-libs/B"),
        Emerge(
            "-p",
            "--newrepo",
            "dev-libs/B",
        ),
        Emerge("-B", "dev-libs/B"),
        Emerge(
            "--oneshot",
            "--usepkg",
            "dev-libs/B",
        ),
        # trigger clean prior to pkg_pretend as in bug #390711
        Ebuild(test_ebuild, "unpack"),
        Emerge("--oneshot", "dev-libs/A"),
        Emerge("--noreplace", "dev-libs/A"),
        Emerge(
            "--config",
            "dev-libs/A",
        ),
        Emerge(
            "--info",
            "dev-libs/A",
            "dev-libs/B",
        ),
        Emerge(
            "--pretend",
            "--depclean",
            "--verbose",
            "dev-libs/B",
        ),
        Emerge("--pretend", "--depclean"),
        Emerge(
            "--depclean",
        ),
        # Test bug #523684, where a file renamed or removed by the
        # admin forces replacement files to be merged with config
        # protection.
        Quickpkg(
            "--include-config",
            "y",
            "dev-libs/A",
            post_command=lambda: _check_number_of_protected_files(
                0, eroot, settings["CONFIG_PROTECT"]
            ),
        ),
        Emerge("--noreplace", "dev-libs/A"),
        Emerge(
            "--usepkgonly",
            "dev-libs/A",
            preparation=lambda: os.unlink(os.path.join(eprefix, "etc", "A-0")),
            post_command=lambda: _check_number_of_protected_files(
                1, eroot, settings["CONFIG_PROTECT"]
            ),
        ),
        Emaint("--check", "all"),
        Emaint("--fix", "all"),
        Fixpackages(),
        Regenworld(),
        Portageq(
            "match",
            eroot,
            "dev-libs/A",
        ),
        Portageq(
            "best_visible",
            eroot,
            "dev-libs/A",
        ),
        Portageq(
            "best_visible",
            eroot,
            "binary",
            "dev-libs/A",
        ),
        Portageq(
            "contents",
            eroot,
            "dev-libs/A-1",
        ),
        Portageq(
            "metadata",
            eroot,
            "ebuild",
            "dev-libs/A-1",
            "EAPI",
            "IUSE",
            "RDEPEND",
        ),
        Portageq(
            "metadata",
            eroot,
            "binary",
            "dev-libs/A-1",
            "EAPI",
            "USE",
            "RDEPEND",
        ),
        Portageq(
            "metadata",
            eroot,
            "installed",
            "dev-libs/A-1",
            "EAPI",
            "USE",
            "RDEPEND",
        ),
        Portageq(
            "owners",
            eroot,
            eroot + "usr",
        ),
        Emerge("-p", eroot + "usr"),
        Emerge(
            "-p",
            "--unmerge",
            "-q",
            eroot + "usr",
        ),
        Emerge(
            "--unmerge",
            "--quiet",
            "dev-libs/A",
        ),
        Emerge(
            "-C",
            "--quiet",
            "dev-libs/B",
        ),
        # autounmask:
        # If EMERGE_DEFAULT_OPTS contains --autounmask=n, then --autounmask
        # must be specified with --autounmask-continue.
        Emerge(
            "--autounmask",
            "--autounmask-continue",
            "dev-libs/C",
            env_mod={"EMERGE_DEFAULT_OPTS": "--autounmask=n"},
        ),
        # Verify that the above --autounmask-continue command caused
        # USE=flag to be applied correctly to dev-libs/D.
        Portageq(
            "match",
            eroot,
            "dev-libs/D[flag]",
        ),
    ]
    test_commands["misc package operations"] = PortageCommandSequence(*abcd_seq)

    cross_prefix_seq = [
        # Test cross-prefix usage, including chpathtool for binpkgs.
        # EAPI 7
        Emerge("dev-libs/C", env_mod={"EPREFIX": cross_prefix}),
        Portageq(
            "has_version", cross_prefix, "dev-libs/C", env_mod={"EPREFIX": cross_prefix}
        ),
        Portageq(
            "has_version", cross_prefix, "dev-libs/D", env_mod={"EPREFIX": cross_prefix}
        ),
        Emerge("dev-libs/D", env_mod={"ROOT": cross_root}),
        Portageq(
            "has_version",
            cross_eroot,
            "dev-libs/D",
        ),
        # EAPI 5
        Emerge("--usepkgonly", "dev-libs/A", env_mod={"EPREFIX": cross_prefix}),
        Portageq(
            "has_version", cross_prefix, "dev-libs/A", env_mod={"EPREFIX": cross_prefix}
        ),
        Portageq(
            "has_version", cross_prefix, "dev-libs/B", env_mod={"EPREFIX": cross_prefix}
        ),
        Emerge("-C", "--quiet", "dev-libs/B", env_mod={"EPREFIX": cross_prefix}),
        Emerge("-C", "--quiet", "dev-libs/A", env_mod={"EPREFIX": cross_prefix}),
        Emerge("dev-libs/A", env_mod={"EPREFIX": cross_prefix}),
        # Test ROOT support
        Emerge("dev-libs/B", env_mod={"ROOT": cross_root}),
        Portageq(
            "has_version",
            cross_eroot,
            "dev-libs/B",
        ),
    ]
    test_commands["misc operations with eprefix"] = PortageCommandSequence(
        *cross_prefix_seq
    )

    # Test binhost support if FETCHCOMMAND is available.
    binrepos_conf_file = os.path.join(os.sep, eprefix, BINREPOS_CONF_FILE)
    binhost_uri = binhost["uri"]
    binhost_dir = binhost["dir"]
    with open(binrepos_conf_file, "w") as f:
        f.write("[test-binhost]\n")
        f.write(f"sync-uri = {binhost_uri}\n")
    fetchcommand = shlex.split(settings["FETCHCOMMAND"])
    fetch_bin = portage.process.find_binary(fetchcommand[0])

    if fetch_bin is None:
        test_commands["binhost emerge"] = Noop()
    else:
        # The next emerge has been added to split this test from the rest:
        make_package = Emerge("-e", "--buildpkg", "dev-libs/A")
        getbinpkgonly = Emerge(
            "-e",
            "--getbinpkgonly",
            "dev-libs/A",
            preparation=lambda: os.rename(pkgdir, binhost_dir),
        )

        # Remove binrepos.conf and test PORTAGE_BINHOST.
        def _rm_pkgdir_and_rm_binrepos_conf_file():
            shutil.rmtree(pkgdir)
            os.unlink(binrepos_conf_file)

        getbinpkgonly_fetchonly = Emerge(
            "-fe",
            "--getbinpkgonly",
            "dev-libs/A",
            env_mod={"PORTAGE_BINHOST": binhost_uri},
            preparation=_rm_pkgdir_and_rm_binrepos_conf_file,
        )

        # Test bug 920537 binrepos.conf with local file src-uri.
        def _rm_pkgdir_and_create_binrepos_conf_with_file_uri():
            shutil.rmtree(pkgdir)
            with open(binrepos_conf_file, "w") as f:
                f.write("[test-binhost]\n")
                f.write(f"sync-uri = file://{binhost_dir}\n")

        getbinpkgonly_file_uri = Emerge(
            "-fe",
            "--getbinpkgonly",
            "dev-libs/A",
            preparation=_rm_pkgdir_and_create_binrepos_conf_with_file_uri,
        )

        fetch_sequence = PortageCommandSequence(
            make_package, getbinpkgonly, getbinpkgonly_fetchonly, getbinpkgonly_file_uri
        )
        test_commands["binhost emerge"] = fetch_sequence
    yield test_commands


@pytest.fixture()
def baseline_command(request, _generate_all_baseline_commands):
    """A fixture that provides the commands to perform a baseline
    functional test of portage. It uses another fixture, namely
    ``_generate_all_baseline_commands``.
    Pytest caches the fixtures and there is a little performance
    improvement if the commands are generated only once..
    """
    return _generate_all_baseline_commands[request.param]



# Copyright 2012-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage import os
from portage.dbapi.porttree import _async_manifest_fetchlist
from portage.dep import _repo_separator
from portage.localization import _
from portage.util._async.AsyncScheduler import AsyncScheduler
from .ManifestTask import ManifestTask


class ManifestScheduler(AsyncScheduler):
    def __init__(
        self,
        portdb,
        cp_iter=None,
        gpg_cmd=None,
        gpg_vars=None,
        force_sign_key=None,
        **kwargs,
    ):
        AsyncScheduler.__init__(self, **kwargs)

        self._portdb = portdb

        if cp_iter is None:
            cp_iter = self._iter_every_cp()
        self._cp_iter = cp_iter
        self._gpg_cmd = gpg_cmd
        self._gpg_vars = gpg_vars
        self._force_sign_key = force_sign_key
        self._task_iter = self._iter_tasks()

    def _next_task(self):
        return next(self._task_iter)

    def _iter_every_cp(self):
        # List categories individually, in order to start yielding quicker,
        # and in order to reduce latency in case of a signal interrupt.
        cp_all = self._portdb.cp_all
        for category in sorted(self._portdb.categories):
            yield from cp_all(categories=(category,))

    def _iter_tasks(self):
        portdb = self._portdb
        distdir = portdb.settings["DISTDIR"]
        disabled_repos = set()

        for cp in self._cp_iter:
            if self._terminated.is_set():
                break
            # We iterate over portdb.porttrees, since it's common to
            # tweak this attribute in order to adjust repo selection.
            for mytree in portdb.porttrees:
                if self._terminated.is_set():
                    break
                repo_config = portdb.repositories.get_repo_for_location(mytree)
                if not repo_config.create_manifest:
                    if repo_config.name not in disabled_repos:
                        disabled_repos.add(repo_config.name)
                        portage.writemsg(
                            _(
                                ">>> Skipping creating Manifest for %s%s%s; "
                                "repository is configured to not use them\n"
                            )
                            % (cp, _repo_separator, repo_config.name),
                            noiselevel=-1,
                        )
                    continue
                cpv_list = portdb.cp_list(cp, mytree=[repo_config.location])
                if not cpv_list:
                    continue

                # Use _async_manifest_fetchlist(max_jobs=1), since we
                # spawn concurrent ManifestTask instances.
                yield ManifestTask(
                    cp=cp,
                    distdir=distdir,
                    fetchlist_dict=_async_manifest_fetchlist(
                        portdb,
                        repo_config,
                        cp,
                        cpv_list=cpv_list,
                        max_jobs=1,
                        loop=self._event_loop,
                    ),
                    repo_config=repo_config,
                    gpg_cmd=self._gpg_cmd,
                    gpg_vars=self._gpg_vars,
                    force_sign_key=self._force_sign_key,
                )

    def _task_exit(self, task):
        if task.returncode != os.EX_OK:
            if not self._terminated_tasks:
                portage.writemsg(
                    "Error processing %s%s%s, continuing...\n"
                    % (task.cp, _repo_separator, task.repo_config.name),
                    noiselevel=-1,
                )

        AsyncScheduler._task_exit(self, task)



# Copyright 2014-2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from ....sync import _SUBMODULE_PATH_MAP

doc = """Check repos.conf settings and sync repositories."""
__doc__ = doc[:]

module_spec = {
    "name": "sync",
    "description": doc,
    "provides": {
        "sync-module": {
            "name": "sync",
            "sourcefile": "sync",
            "class": "SyncRepos",
            "description": doc,
            "functions": ["allrepos", "auto", "repo"],
            "func_desc": {
                "repo": {
                    "short": "-r",
                    "long": "--repo",
                    "help": "(sync module only): -r, --repo  Sync the specified repo",
                    "status": "Syncing %s",
                    "action": "store",
                    "func": "repo",
                },
                "allrepos": {
                    "short": "-A",
                    "long": "--allrepos",
                    "help": "(sync module only): -A, --allrepos  Sync all repos that have a sync-url defined",
                    "status": "Syncing %s",
                    "action": "store_true",
                    "dest": "allrepos",
                    "func": "all_repos",
                },
                "auto": {
                    "short": "-a",
                    "long": "--auto",
                    "help": "(sync module only): -a, --auto  Sync auto-sync enabled repos only",
                    "status": "Syncing %s",
                    "action": "store_true",
                    "dest": "auto",
                    "func": "auto_sync",
                },
            },
            "opt_desc": {
                "sync-submodule": {
                    "long": "--sync-submodule",
                    "help": (
                        "(sync module only): Restrict sync "
                        "to the specified submodule(s)"
                    ),
                    "choices": tuple(_SUBMODULE_PATH_MAP),
                    "action": "append",
                    "dest": "sync_submodule",
                },
            },
        }
    },
}



# Copyright 2010-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

__all__ = ["prepare_build_dirs"]

import errno
import gzip
import stat
import time
import pwd

import portage
from portage import os, shutil, _encodings, _unicode_encode, _unicode_decode
from portage.data import portage_gid, portage_uid, secpass
from portage.exception import (
    DirectoryNotFound,
    FileNotFound,
    OperationNotPermitted,
    PermissionDenied,
    PortageException,
)
from portage.localization import _
from portage.output import colorize
from portage.util import (
    apply_recursive_permissions,
    apply_secpass_permissions,
    ensure_dirs,
    normalize_path,
    writemsg,
)
from portage.util.install_mask import _raise_exc


def prepare_build_dirs(myroot=None, settings=None, cleanup=False):
    """
    The myroot parameter is ignored.
    """
    myroot = None

    if settings is None:
        raise TypeError("settings argument is required")

    mysettings = settings
    clean_dirs = [mysettings["HOME"]]

    # We enable cleanup when we want to make sure old cruft (such as the old
    # environment) doesn't interfere with the current phase.
    if cleanup and "keeptemp" not in mysettings.features:
        clean_dirs.append(mysettings["T"])

    for clean_dir in clean_dirs:
        try:
            shutil.rmtree(clean_dir)
        except OSError as oe:
            if errno.ENOENT == oe.errno:
                pass
            elif errno.EPERM == oe.errno:
                writemsg(f"{oe}\n", noiselevel=-1)
                writemsg(
                    _("Operation Not Permitted: rmtree('%s')\n") % clean_dir,
                    noiselevel=-1,
                )
                return 1
            else:
                # Wrap with PermissionDenied if appropriate, so that callers
                # display a short error message without a traceback.
                _raise_exc(oe)

    def makedirs(dir_path):
        try:
            os.makedirs(dir_path)
        except OSError as oe:
            if errno.EEXIST == oe.errno:
                pass
            elif errno.EPERM == oe.errno:
                writemsg(f"{oe}\n", noiselevel=-1)
                writemsg(
                    _("Operation Not Permitted: makedirs('%s')\n") % dir_path,
                    noiselevel=-1,
                )
                return False
            else:
                raise
        return True

    mysettings["PKG_LOGDIR"] = os.path.join(mysettings["T"], "logging")

    mydirs = [os.path.dirname(mysettings["PORTAGE_BUILDDIR"])]
    mydirs.append(os.path.dirname(mydirs[-1]))

    try:
        for mydir in mydirs:
            ensure_dirs(mydir)
            try:
                apply_secpass_permissions(
                    mydir, gid=portage_gid, uid=portage_uid, mode=0o700, mask=0
                )
            except PortageException:
                if not os.path.isdir(mydir):
                    raise
        for dir_key in ("HOME", "PKG_LOGDIR", "T"):
            ensure_dirs(mysettings[dir_key], mode=0o755)
            apply_secpass_permissions(
                mysettings[dir_key], uid=portage_uid, gid=portage_gid
            )
        # The setgid bit prevents a lockfile group permission race for bug #468990.
        ipc_kwargs = {}
        if portage.data.secpass >= 1:
            ipc_kwargs["gid"] = portage_gid
            ipc_kwargs["mode"] = 0o2770
        ensure_dirs(
            os.path.join(mysettings["PORTAGE_BUILDDIR"], ".ipc"),
            **ipc_kwargs,
        )
    except PermissionDenied as e:
        writemsg(_("Permission Denied: %s\n") % str(e), noiselevel=-1)
        return 1
    except OperationNotPermitted as e:
        writemsg(_("Operation Not Permitted: %s\n") % str(e), noiselevel=-1)
        return 1
    except FileNotFound as e:
        writemsg(_("File Not Found: '%s'\n") % str(e), noiselevel=-1)
        return 1

    # Reset state for things like noauto and keepwork in FEATURES.
    for x in (".die_hooks",):
        try:
            os.unlink(os.path.join(mysettings["PORTAGE_BUILDDIR"], x))
        except OSError:
            pass

    _prepare_workdir(mysettings)
    if mysettings.get("EBUILD_PHASE") not in ("info", "fetch", "pretend"):
        # Avoid spurious permissions adjustments when fetching with
        # a temporary PORTAGE_TMPDIR setting (for fetchonly).
        _prepare_features_dirs(mysettings)
        # Support for home-dir-template-copy FEATURE:
        if "home-dir-template-copy" in settings.features:
            portage_username = mysettings.get("PORTAGE_USERNAME", "portage")
            home_template_dir = pwd.getpwnam(portage_username).pw_dir
            build_env_home_dir = mysettings["HOME"]
            # Sanity checks on above values.
            if not os.path.exists(home_template_dir):
                writemsg(
                    f"FEATURES home-dir-template-copy enabled but specified Linux home directory {home_template_dir} does not exist.",
                    noiselevel=-1,
                )
                return 1
            if not os.path.exists(build_env_home_dir):
                writemsg(
                    f"FEATURES home-dir-template-copy enabled but build HOME directory {build_env_home_dir} does not exist.",
                    noiselevel=-1,
                )
                return 1
            shutil.copytree(
                home_template_dir, build_env_home_dir, symlinks=True, dirs_exist_ok=True
            )


def _adjust_perms_msg(settings, msg):
    def write(msg):
        writemsg(msg, noiselevel=-1)

    background = settings.get("PORTAGE_BACKGROUND") == "1"
    log_path = settings.get("PORTAGE_LOG_FILE")
    log_file = None
    log_file_real = None

    if background and log_path is not None:
        try:
            log_file = open(
                _unicode_encode(log_path, encoding=_encodings["fs"], errors="strict"),
                mode="ab",
            )
            log_file_real = log_file
        except OSError:

            def write(msg):
                pass

        else:
            if log_path.endswith(".gz"):
                log_file = gzip.GzipFile(filename="", mode="ab", fileobj=log_file)

            def write(msg):
                log_file.write(_unicode_encode(msg))
                log_file.flush()

    try:
        write(msg)
    finally:
        if log_file is not None:
            log_file.close()
            if log_file_real is not log_file:
                log_file_real.close()


def _prepare_features_dirs(mysettings):
    # Use default ABI libdir in accordance with bug #355283.
    libdir = None
    default_abi = mysettings.get("DEFAULT_ABI")
    if default_abi:
        libdir = mysettings.get("LIBDIR_" + default_abi)
    if not libdir:
        libdir = "lib"

    features_dirs = {
        "ccache": {
            "basedir_var": "CCACHE_DIR",
            "default_dir": os.path.join(mysettings["PORTAGE_TMPDIR"], "ccache"),
            "always_recurse": False,
        },
        "distcc": {
            "basedir_var": "DISTCC_DIR",
            "default_dir": os.path.join(mysettings["BUILD_PREFIX"], ".distcc"),
            "subdirs": ("lock", "state"),
            "always_recurse": True,
        },
    }
    dirmode = 0o2070
    filemode = 0o60
    modemask = 0o2
    restrict = mysettings.get("PORTAGE_RESTRICT", "").split()
    droppriv = (
        secpass >= 2
        and "userpriv" in mysettings.features
        and "userpriv" not in restrict
    )
    for myfeature, kwargs in features_dirs.items():
        if myfeature in mysettings.features:
            failure = False
            basedir = mysettings.get(kwargs["basedir_var"])
            if basedir is None or not basedir.strip():
                basedir = kwargs["default_dir"]
                mysettings[kwargs["basedir_var"]] = basedir
            try:
                mydirs = [mysettings[kwargs["basedir_var"]]]
                if "subdirs" in kwargs:
                    for subdir in kwargs["subdirs"]:
                        mydirs.append(os.path.join(basedir, subdir))
                for mydir in mydirs:
                    modified = ensure_dirs(mydir)
                    # Generally, we only want to apply permissions for
                    # initial creation.  Otherwise, we don't know exactly what
                    # permissions the user wants, so should leave them as-is.
                    droppriv_fix = False
                    if droppriv:
                        st = os.stat(mydir)
                        if st.st_gid != portage_gid or not dirmode == (
                            stat.S_IMODE(st.st_mode) & dirmode
                        ):
                            droppriv_fix = True
                        if not droppriv_fix:
                            # Check permissions of files in the directory.
                            for filename in os.listdir(mydir):
                                try:
                                    subdir_st = os.lstat(os.path.join(mydir, filename))
                                except OSError:
                                    continue
                                if subdir_st.st_gid != portage_gid or (
                                    stat.S_ISDIR(subdir_st.st_mode)
                                    and not dirmode
                                    == (stat.S_IMODE(subdir_st.st_mode) & dirmode)
                                ):
                                    droppriv_fix = True
                                    break

                    if droppriv_fix:
                        _adjust_perms_msg(
                            mysettings,
                            colorize("WARN", " * ")
                            + _(
                                "Adjusting permissions " "for FEATURES=userpriv: '%s'\n"
                            )
                            % mydir,
                        )
                    elif modified:
                        _adjust_perms_msg(
                            mysettings,
                            colorize("WARN", " * ")
                            + _("Adjusting permissions " "for FEATURES=%s: '%s'\n")
                            % (myfeature, mydir),
                        )

                    if modified or kwargs["always_recurse"] or droppriv_fix:

                        def onerror(e):
                            raise  # The feature is disabled if a single error
                            # occurs during permissions adjustment.

                        if not apply_recursive_permissions(
                            mydir,
                            gid=portage_gid,
                            dirmode=dirmode,
                            dirmask=modemask,
                            filemode=filemode,
                            filemask=modemask,
                            onerror=onerror,
                        ):
                            raise OperationNotPermitted(
                                _(
                                    "Failed to apply recursive permissions for the portage group."
                                )
                            )

            except DirectoryNotFound as e:
                failure = True
                writemsg(
                    _("\n!!! Directory does not exist: '%s'\n") % (e,), noiselevel=-1
                )
                writemsg(_("!!! Disabled FEATURES='%s'\n") % myfeature, noiselevel=-1)

            except PortageException as e:
                failure = True
                writemsg(f"\n!!! {str(e)}\n", noiselevel=-1)
                writemsg(
                    _("!!! Failed resetting perms on %s='%s'\n")
                    % (kwargs["basedir_var"], basedir),
                    noiselevel=-1,
                )
                writemsg(_("!!! Disabled FEATURES='%s'\n") % myfeature, noiselevel=-1)

            if failure:
                mysettings.features.remove(myfeature)
                time.sleep(5)


def _prepare_workdir(mysettings):
    workdir_mode = 0o700
    try:
        mode = mysettings["PORTAGE_WORKDIR_MODE"]
        if mode.isdigit():
            parsed_mode = int(mode, 8)
        elif mode == "":
            raise KeyError()
        else:
            raise ValueError()
        if parsed_mode & 0o7777 != parsed_mode:
            raise ValueError(f"Invalid file mode: {mode}")
        else:
            workdir_mode = parsed_mode
    except KeyError as e:
        writemsg(
            _("!!! PORTAGE_WORKDIR_MODE is unset, using %s.\n") % oct(workdir_mode)
        )
    except ValueError as e:
        if len(str(e)) > 0:
            writemsg(f"{e}\n")
        writemsg(
            _("!!! Unable to parse PORTAGE_WORKDIR_MODE='%s', using %s.\n")
            % (mysettings["PORTAGE_WORKDIR_MODE"], oct(workdir_mode))
        )
    mysettings["PORTAGE_WORKDIR_MODE"] = oct(workdir_mode).replace("o", "")

    permissions = {"mode": workdir_mode}
    if portage.data.secpass >= 2:
        permissions["uid"] = portage_uid
    if portage.data.secpass >= 1:
        permissions["gid"] = portage_gid

    # Apply PORTAGE_WORKDIR_MODE to PORTAGE_BUILDDIR, since the child
    # directory ${D} and its children may have vulnerable permissions
    # as reported in bug 692492.
    ensure_dirs(mysettings["PORTAGE_BUILDDIR"], **permissions)
    ensure_dirs(mysettings["WORKDIR"], **permissions)

    if mysettings.get("PORTAGE_LOGDIR", "") == "":
        while "PORTAGE_LOGDIR" in mysettings:
            del mysettings["PORTAGE_LOGDIR"]
    if "PORTAGE_LOGDIR" in mysettings:
        try:
            modified = ensure_dirs(mysettings["PORTAGE_LOGDIR"])
            if modified:
                # Only initialize group/mode if the directory doesn't
                # exist, so that we don't override permissions if they
                # were previously set by the administrator.
                # NOTE: These permissions should be compatible with our
                # default logrotate config as discussed in bug 374287.
                apply_secpass_permissions(
                    mysettings["PORTAGE_LOGDIR"],
                    uid=portage_uid,
                    gid=portage_gid,
                    mode=0o2770,
                )
        except PortageException as e:
            writemsg(f"!!! {str(e)}\n", noiselevel=-1)
            writemsg(
                _("!!! Permission issues with PORTAGE_LOGDIR='%s'\n")
                % mysettings["PORTAGE_LOGDIR"],
                noiselevel=-1,
            )
            writemsg(_("!!! Disabling logging.\n"), noiselevel=-1)
            while "PORTAGE_LOGDIR" in mysettings:
                del mysettings["PORTAGE_LOGDIR"]

    compress_log_ext = ""
    if "compress-build-logs" in mysettings.features:
        compress_log_ext = ".gz"

    logdir_subdir_ok = False
    if "PORTAGE_LOGDIR" in mysettings and os.access(
        mysettings["PORTAGE_LOGDIR"], os.W_OK
    ):
        logdir = normalize_path(mysettings["PORTAGE_LOGDIR"])
        logid_path = os.path.join(mysettings["PORTAGE_BUILDDIR"], ".logid")
        if not os.path.exists(logid_path):
            open(_unicode_encode(logid_path), "w").close()
        logid_time = _unicode_decode(
            time.strftime("%Y%m%d-%H%M%S", time.gmtime(os.stat(logid_path).st_mtime)),
            encoding=_encodings["content"],
            errors="replace",
        )

        # The separator used between the individual name components of the log file.
        sep = mysettings.get("PORTAGE_LOG_FILE_SEP", ":")

        if "split-log" in mysettings.features:
            log_subdir = os.path.join(logdir, "build", mysettings["CATEGORY"])
            mysettings["PORTAGE_LOG_FILE"] = os.path.join(
                log_subdir,
                f"{mysettings['PF']}{sep}{logid_time}.log{compress_log_ext}",
            )
        else:
            log_subdir = logdir
            mysettings["PORTAGE_LOG_FILE"] = os.path.join(
                logdir,
                f"{mysettings['CATEGORY']}{sep}{mysettings['PF']}{sep}{logid_time}.log{compress_log_ext}",
            )

        if log_subdir is logdir:
            logdir_subdir_ok = True
        else:
            try:
                _ensure_log_subdirs(logdir, log_subdir)
            except PortageException as e:
                writemsg(f"!!! {e}\n", noiselevel=-1)

            if os.access(log_subdir, os.W_OK):
                logdir_subdir_ok = True
            else:
                writemsg(
                    f"!!! {_('Permission Denied')}: {log_subdir}\n",
                    noiselevel=-1,
                )

    tmpdir_log_path = os.path.join(mysettings["T"], f"build.log{compress_log_ext}")
    if not logdir_subdir_ok:
        # NOTE: When sesandbox is enabled, the local SELinux security policies
        # may not allow output to be piped out of the sesandbox domain. The
        # current policy will allow it to work when a pty is available, but
        # not through a normal pipe. See bug #162404.
        mysettings["PORTAGE_LOG_FILE"] = tmpdir_log_path
    else:
        # Create a symlink from tmpdir_log_path to PORTAGE_LOG_FILE, as
        # requested in bug #412865.
        make_new_symlink = False
        try:
            target = os.readlink(tmpdir_log_path)
        except OSError:
            make_new_symlink = True
        else:
            if target != mysettings["PORTAGE_LOG_FILE"]:
                make_new_symlink = True
        if make_new_symlink:
            try:
                os.unlink(tmpdir_log_path)
            except OSError:
                pass
            os.symlink(mysettings["PORTAGE_LOG_FILE"], tmpdir_log_path)


def _ensure_log_subdirs(logdir, subdir):
    """
    This assumes that logdir exists, and creates subdirectories down
    to subdir as necessary. The gid of logdir is copied to all
    subdirectories, along with 0x2070 mode bits if present. Both logdir
    and subdir are assumed to be normalized absolute paths.
    """
    st = os.stat(logdir)
    uid = -1
    gid = st.st_gid
    grp_mode = 0o2070 & st.st_mode

    # If logdir is writable by the portage group but its uid
    # is not portage_uid, then set the uid to portage_uid if
    # we have privileges to do so, for compatibility with our
    # default logrotate config (see bug 378451). With the
    # "su portage portage" directive and logrotate-3.8.0,
    # logrotate's chown call during the compression phase will
    # only succeed if the log file's uid is portage_uid.
    if grp_mode and gid == portage_gid and portage.data.secpass >= 2:
        uid = portage_uid
        if st.st_uid != portage_uid:
            ensure_dirs(logdir, uid=uid)

    logdir_split_len = len(logdir.split(os.sep))
    subdir_split = subdir.split(os.sep)[logdir_split_len:]
    subdir_split.reverse()
    current = logdir
    while subdir_split:
        current = os.path.join(current, subdir_split.pop())
        ensure_dirs(current, uid=uid, gid=gid, mode=grp_mode, mask=0)


def _prepare_fake_filesdir(settings):
    real_filesdir = settings["O"] + "/files"
    symlink_path = settings["FILESDIR"]

    try:
        link_target = os.readlink(symlink_path)
    except OSError:
        os.symlink(real_filesdir, symlink_path)
    else:
        if link_target != real_filesdir:
            os.unlink(symlink_path)
            os.symlink(real_filesdir, symlink_path)


def _prepare_fake_distdir(settings, alist):
    orig_distdir = settings["DISTDIR"]
    edpath = os.path.join(settings["PORTAGE_BUILDDIR"], "distdir")
    portage.util.ensure_dirs(edpath, gid=portage_gid, mode=0o755)

    # Remove any unexpected files or directories.
    for x in os.listdir(edpath):
        symlink_path = os.path.join(edpath, x)
        st = os.lstat(symlink_path)
        if x in alist and stat.S_ISLNK(st.st_mode):
            continue
        if stat.S_ISDIR(st.st_mode):
            shutil.rmtree(symlink_path)
        else:
            os.unlink(symlink_path)

    # Check for existing symlinks and recreate if necessary.
    for x in alist:
        symlink_path = os.path.join(edpath, x)
        target = os.path.join(orig_distdir, x)
        try:
            link_target = os.readlink(symlink_path)
        except OSError:
            os.symlink(target, symlink_path)
        else:
            if link_target != target:
                os.unlink(symlink_path)
                os.symlink(target, symlink_path)



# Copyright 2010-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import asyncio
import platform
import pty
import termios
from typing import Optional, Union

from portage import os
from portage.output import get_term_size, set_term_size
from portage.util import writemsg

# Disable the use of openpty on Solaris as it seems Python's openpty
# implementation doesn't play nice on Solaris with Portage's
# behaviour causing hangs/deadlocks.
# Additional note for the future: on Interix, pipes do NOT work, so
# _disable_openpty on Interix must *never* be True
_disable_openpty = platform.system() in ("SunOS",)

_fbsd_test_pty = platform.system() == "FreeBSD"


def _create_pty_or_pipe(
    copy_term_size: Optional[int] = None,
) -> tuple[Union[asyncio.Future, bool], int, int]:
    """
    Try to create a pty and if then fails then create a normal
    pipe instead. If a Future is returned for pty_ready, then the
    caller should wait for it (which comes from set_term_size
    because it spawns stty).

    @param copy_term_size: If a tty file descriptor is given
            then the term size will be copied to the pty.
    @type copy_term_size: int
    @rtype: tuple
    @return: A tuple of (pty_ready, master_fd, slave_fd) where
            pty_ready is asyncio.Future or True if a pty was successfully allocated, and
            False if a normal pipe was allocated.
    """

    got_pty = False

    global _disable_openpty, _fbsd_test_pty

    if _fbsd_test_pty and not _disable_openpty:
        # Test for python openpty breakage after freebsd7 to freebsd8
        # upgrade, which results in a 'Function not implemented' error
        # and the process being killed.
        pid = os.fork()
        if pid == 0:
            pty.openpty()
            os._exit(os.EX_OK)
        pid, status = os.waitpid(pid, 0)
        if (status & 0xFF) == 140:
            _disable_openpty = True
        _fbsd_test_pty = False

    if _disable_openpty:
        master_fd, slave_fd = os.pipe()
    else:
        try:
            master_fd, slave_fd = pty.openpty()
            got_pty = True
        except OSError as e:
            _disable_openpty = True
            writemsg(f"openpty failed: '{str(e)}'\n", noiselevel=-1)
            del e
            master_fd, slave_fd = os.pipe()

    if got_pty:
        # Disable post-processing of output since otherwise weird
        # things like \n -> \r\n transformations may occur.
        mode = termios.tcgetattr(slave_fd)
        mode[1] &= ~termios.OPOST
        termios.tcsetattr(slave_fd, termios.TCSANOW, mode)

    pty_ready = None
    if got_pty and copy_term_size is not None and os.isatty(copy_term_size):
        rows, columns = get_term_size()
        pty_ready = set_term_size(rows, columns, slave_fd)

    # The future only exists when got_pty is True, so we can
    # return the future in lieu of got_pty when it exists.
    return (got_pty if pty_ready is None else pty_ready, master_fd, slave_fd)



# portage: Constants
# Copyright 1998-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import os

from portage import installation

# ===========================================================================
# START OF CONSTANTS -- START OF CONSTANTS -- START OF CONSTANTS -- START OF
# ===========================================================================

# There are two types of variables here which can easily be confused,
# resulting in arbitrary bugs, mainly exposed with an offset
# installation (Prefix).  The two types relate to the usage of
# config_root or target_root.
# The first, config_root (PORTAGE_CONFIGROOT), can be a path somewhere,
# from which all derived paths need to be relative (e.g.
# USER_CONFIG_PATH) without EPREFIX prepended in Prefix.  This means
# config_root can for instance be set to "$HOME/my/config".  Obviously,
# in such case it is not appropriate to prepend EPREFIX to derived
# constants.  The default value of config_root is EPREFIX (in non-Prefix
# the empty string) -- overriding the value loses the EPREFIX as one
# would expect.
# Second there is target_root (ROOT) which is used to install somewhere
# completely else, in Prefix of limited use.  Because this is an offset
# always given, the EPREFIX should always be applied in it, hence the
# code always prefixes them with EROOT.
# The variables in this file are grouped by config_root, target_root.

# variables used with config_root (these need to be relative)
USER_CONFIG_PATH = "etc/portage"
BINREPOS_CONF_FILE = f"{USER_CONFIG_PATH}/binrepos.conf"
MAKE_CONF_FILE = f"{USER_CONFIG_PATH}/make.conf"
MODULES_FILE_PATH = f"{USER_CONFIG_PATH}/modules"
CUSTOM_PROFILE_PATH = f"{USER_CONFIG_PATH}/profile"
USER_VIRTUALS_FILE = f"{USER_CONFIG_PATH}/virtuals"
EBUILD_SH_ENV_FILE = f"{USER_CONFIG_PATH}/bashrc"
EBUILD_SH_ENV_DIR = f"{USER_CONFIG_PATH}/env"
CUSTOM_MIRRORS_FILE = f"{USER_CONFIG_PATH}/mirrors"
COLOR_MAP_FILE = f"{USER_CONFIG_PATH}/color.map"
PROFILE_PATH = f"{USER_CONFIG_PATH}/make.profile"
MAKE_DEFAULTS_FILE = f"{PROFILE_PATH}/make.defaults"  # FIXME: not used
DEPRECATED_PROFILE_FILE = f"{PROFILE_PATH}/deprecated"

# variables used with targetroot (these need to be absolute, but not
# have a leading '/' since they are used directly with os.path.join on EROOT)
VDB_PATH = "var/db/pkg"
CACHE_PATH = "var/cache/edb"
PRIVATE_PATH = "var/lib/portage"
WORLD_FILE = f"{PRIVATE_PATH}/world"
WORLD_SETS_FILE = f"{PRIVATE_PATH}/world_sets"
CONFIG_MEMORY_FILE = f"{PRIVATE_PATH}/config"
REPO_REVISIONS = f"{PRIVATE_PATH}/repo_revisions"
NEWS_LIB_PATH = "var/lib/gentoo"

# these variables get EPREFIX prepended automagically when they are
# translated into their lowercase variants
DEPCACHE_PATH = f"/{CACHE_PATH}/dep"

if installation.TYPE == installation.TYPES.MODULE:
    GLOBAL_CONFIG_PATH = "/share/portage/config"
else:
    GLOBAL_CONFIG_PATH = "/usr/share/portage/config"

# these variables are not used with target_root or config_root
# NOTE: Use realpath(__file__) so that python module symlinks in site-packages
# are followed back to the real location of the whole portage installation.
if installation.TYPE == installation.TYPES.SYSTEM:
    PORTAGE_BASE_PATH = """@PORTAGE_BASE_PATH@"""
elif installation.TYPE == installation.TYPES.MODULE:
    PORTAGE_BASE_PATH = os.path.join(
        os.path.realpath(__import__("sys").prefix), "lib/portage"
    )
else:
    PORTAGE_BASE_PATH = os.path.join(
        os.sep, *os.path.realpath(__file__).split(os.sep)[:-3]
    )

if installation.TYPE == installation.TYPES.SYSTEM:
    PORTAGE_BIN_PATH = """@PORTAGE_BIN_PATH@"""
else:
    PORTAGE_BIN_PATH = f"{PORTAGE_BASE_PATH}/bin"

# The EPREFIX for the current install is hardcoded here, but access to this
# constant should be minimal, in favor of access via the EPREFIX setting of
# a config instance (since it's possible to contruct a config instance with
# a different EPREFIX). Therefore, the EPREFIX constant should *NOT* be used
# in the definition of any other constants within this file.
if installation.TYPE == installation.TYPES.SYSTEM:
    EPREFIX = BINARY_PREFIX = "@EPREFIX@"
elif installation.TYPE == installation.TYPES.MODULE:
    EPREFIX = __import__("sys").prefix
    BINARY_PREFIX = ""
else:
    EPREFIX = BINARY_PREFIX = ""

PORTAGE_PYM_PATH = os.path.realpath(os.path.join(__file__, "../.."))
LOCALE_DATA_PATH = f"{PORTAGE_BASE_PATH}/locale"  # FIXME: not used
EBUILD_SH_BINARY = f"{PORTAGE_BIN_PATH}/ebuild.sh"
MISC_SH_BINARY = f"{PORTAGE_BIN_PATH}/misc-functions.sh"
SANDBOX_BINARY = f"{BINARY_PREFIX}/usr/bin/sandbox"
FAKEROOT_BINARY = f"{BINARY_PREFIX}/usr/bin/fakeroot"
BASH_BINARY = f"{BINARY_PREFIX}/bin/bash"
MOVE_BINARY = f"{BINARY_PREFIX}/bin/mv"
PRELINK_BINARY = f"{BINARY_PREFIX}/usr/sbin/prelink"

INVALID_ENV_FILE = "/etc/spork/is/not/valid/profile.env"
MERGING_IDENTIFIER = "-MERGING-"
REPO_NAME_FILE = "repo_name"
REPO_NAME_LOC = f"profiles/{REPO_NAME_FILE}"

PORTAGE_PACKAGE_ATOM = "sys-apps/portage"
LIBC_PACKAGE_ATOM = "virtual/libc"
OS_HEADERS_PACKAGE_ATOM = "virtual/os-headers"
CVS_PACKAGE_ATOM = "dev-vcs/cvs"
GIT_PACKAGE_ATOM = "dev-vcs/git"
HG_PACKAGE_ATOM = "dev-vcs/mercurial"
RSYNC_PACKAGE_ATOM = "net-misc/rsync"

INCREMENTALS = (
    "ACCEPT_KEYWORDS",
    "CONFIG_PROTECT",
    "CONFIG_PROTECT_MASK",
    "ENV_UNSET",
    "FEATURES",
    "IUSE_IMPLICIT",
    "PRELINK_PATH",
    "PRELINK_PATH_MASK",
    "PROFILE_ONLY_VARIABLES",
    "USE",
    "USE_EXPAND",
    "USE_EXPAND_HIDDEN",
    "USE_EXPAND_IMPLICIT",
    "USE_EXPAND_UNPREFIXED",
)
EBUILD_PHASES = (
    "pretend",
    "setup",
    "unpack",
    "prepare",
    "configure",
    "compile",
    "test",
    "install",
    "package",
    "instprep",
    "preinst",
    "postinst",
    "prerm",
    "postrm",
    "nofetch",
    "config",
    "info",
    "other",
)
SUPPORTED_FEATURES = frozenset(
    (
        "assume-digests",
        "binpkg-docompress",
        "binpkg-dostrip",
        "binpkg-ignore-signature",
        "binpkg-logs",
        "binpkg-multi-instance",
        "binpkg-request-signature",
        "binpkg-signing",
        "buildpkg",
        "buildpkg-live",
        "buildsyspkg",
        "candy",
        "case-insensitive-fs",
        "ccache",
        "chflags",
        "clean-logs",
        "collision-protect",
        "compress-build-logs",
        "compressdebug",
        "compress-index",
        "config-protect-if-modified",
        "dedupdebug",
        "digest",
        "distcc",
        "distlocks",
        "downgrade-backup",
        "ebuild-locks",
        "export-pms-vars",
        "fail-clean",
        "fakeroot",
        "fixlafiles",
        "force-mirror",
        "getbinpkg",
        "gpg-keepalive",
        "home-dir-template-copy",
        "icecream",
        "installsources",
        "ipc-sandbox",
        "keeptemp",
        "keepwork",
        "lmirror",
        "merge-sync",
        "merge-wait",
        "metadata-transfer",
        "mirror",
        "mount-sandbox",
        "multilib-strict",
        "network-sandbox",
        "network-sandbox-proxy",
        "news",
        "noauto",
        "noclean",
        "nodoc",
        "noinfo",
        "noman",
        "nostrip",
        "notitles",
        "parallel-fetch",
        "parallel-install",
        "pid-sandbox",
        "pkgdir-index-trusted",
        "prelink-checksums",
        "preserve-libs",
        "protect-owned",
        "python-trace",
        "qa-unresolved-soname-deps",
        "sandbox",
        "selinux",
        "sesandbox",
        "sfperms",
        "sign",
        "skiprocheck",
        "splitdebug",
        "split-elog",
        "split-log",
        "strict",
        "strict-keepdir",
        "stricter",
        "suidctl",
        "test",
        "test-fail-continue",
        "unknown-features-filter",
        "unknown-features-warn",
        "unmerge-backup",
        "unmerge-logs",
        "unmerge-orphans",
        "unprivileged",
        "userfetch",
        "userpriv",
        "usersandbox",
        "usersync",
        "warn-on-large-env",
        "webrsync-gpg",
        "xattr",
    )
)

EAPI = 8

HASHING_BLOCKSIZE = 32768

MANIFEST2_HASH_DEFAULTS = frozenset(("BLAKE2B", "SHA512"))
MANIFEST2_HASH_DEFAULT = "BLAKE2B"

MANIFEST2_IDENTIFIERS = ("AUX", "MISC", "DIST", "EBUILD")

# Redefine EPREFIX from the environment if set
if "PORTAGE_OVERRIDE_EPREFIX" in os.environ:
    EPREFIX = os.environ["PORTAGE_OVERRIDE_EPREFIX"]
    if EPREFIX:
        EPREFIX = os.path.normpath(EPREFIX)
        if EPREFIX == os.sep:
            EPREFIX = ""

VCS_DIRS = ("CVS", "RCS", "SCCS", ".bzr", ".git", ".hg", ".svn")

# List of known live eclasses. Keep it in sync with cnf/sets/portage.conf
LIVE_ECLASSES = frozenset(
    (
        "bzr",
        "cvs",
        "darcs",
        "git-2",
        "git-r3",
        "golang-vcs",
        "mercurial",
        "subversion",
    )
)

SUPPORTED_BINPKG_FORMATS = ("tar", "rpm")

SUPPORTED_GENTOO_BINPKG_FORMATS = ("xpak", "gpkg")

SUPPORTED_XPAK_EXTENSIONS = (".tbz2", ".xpak")
SUPPORTED_GPKG_EXTENSIONS = (".gpkg.tar",)

# Time formats used in various places like metadata.chk.
TIMESTAMP_FORMAT = "%a, %d %b %Y %H:%M:%S +0000"  # to be used with time.gmtime()

# Top-level names of Python packages installed by Portage.
PORTAGE_PYM_PACKAGES = ("_emerge", "portage")

RETURNCODE_POSTINST_FAILURE = 5

# ===========================================================================
# END OF CONSTANTS -- END OF CONSTANTS -- END OF CONSTANTS -- END OF CONSTANT
# ===========================================================================

# Private constants for use in conditional code in order to minimize the diff
# between branches.
_DEPCLEAN_LIB_CHECK_DEFAULT = True
_ENABLE_SET_CONFIG = True



# Copyright 2001-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import shlex
import subprocess
import sys
import threading

from portage import os
from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.exception import GPGException
from portage.output import colorize
from portage.util import varexpand, writemsg, writemsg_stdout


class GPG:
    """
    Unlock GPG, must call dircetly from main program for get correct TTY
    """

    def __init__(self, settings):
        """
        Portage settings are needed to run GPG unlock command.
        """
        self.settings = settings
        self.thread = None
        self._terminated = None
        self.GPG_signing_base_command = self.settings.get(
            "BINPKG_GPG_SIGNING_BASE_COMMAND"
        )
        self.digest_algo = self.settings.get("BINPKG_GPG_SIGNING_DIGEST")
        self.signing_gpg_home = self.settings.get("BINPKG_GPG_SIGNING_GPG_HOME")
        self.signing_gpg_key = self.settings.get("BINPKG_GPG_SIGNING_KEY")
        self.GPG_unlock_command = self.GPG_signing_base_command.replace(
            "[PORTAGE_CONFIG]",
            f"--homedir {self.signing_gpg_home} "
            f"--digest-algo {self.digest_algo} "
            f"--local-user {self.signing_gpg_key} "
            "--output - /dev/null",
        )

        if "gpg-keepalive" in self.settings.features:
            self.keepalive = True
        else:
            self.keepalive = False

    def unlock(self):
        """
        Set GPG_TTY and run GPG unlock command.
        If gpg-keepalive is set, start keepalive thread.
        """
        if self.GPG_unlock_command and (
            self.settings.get("BINPKG_FORMAT", SUPPORTED_GENTOO_BINPKG_FORMATS[0])
            == "gpkg"
        ):
            try:
                os.environ["GPG_TTY"] = os.ttyname(sys.stdout.fileno())
            except OSError as e:
                # When run with no input/output tty, this will fail.
                # However, if the password is given by command,
                # GPG does not need to ask password, so can be ignored.
                writemsg(f"{colorize('WARN', str(e))}\n")

            cmd = shlex.split(varexpand(self.GPG_unlock_command, mydict=self.settings))
            return_code = subprocess.Popen(cmd, stdout=subprocess.DEVNULL).wait()

            if return_code == os.EX_OK:
                writemsg_stdout(f"{colorize('GOOD', 'unlocked')}\n")
                sys.stdout.flush()
            else:
                raise GPGException("GPG unlock failed")

            if self.keepalive:
                self.GPG_unlock_command = shlex.split(
                    varexpand(self.GPG_unlock_command, mydict=self.settings)
                )
                self._terminated = threading.Event()
                self.thread = threading.Thread(target=self.gpg_keepalive, daemon=True)
                self.thread.start()

    def stop(self):
        """
        Stop keepalive thread.
        """
        if self.thread is not None:
            self._terminated.set()

    def gpg_keepalive(self):
        """
        Call GPG unlock command every 5 mins to avoid the passphrase expired.
        """
        count = 0
        while not self._terminated.is_set():
            if count < 5:
                if self._terminated.wait(60):
                    break
                count += 1
                continue
            else:
                count = 0

            proc = subprocess.Popen(
                self.GPG_unlock_command,
                stdin=subprocess.DEVNULL,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.STDOUT,
            )
            if proc.wait() != os.EX_OK and not self._terminated.is_set():
                raise GPGException("GPG keepalive failed")



# Copyright 2025 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import pytest

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class BootstrapChainTestCase(TestCase):
    @pytest.mark.xfail(reason="bug #950310")
    def testBootstrapChain(self):
        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "8",
                "SLOT": "1",
            },
            "dev-libs/A-2": {
                "EAPI": "8",
                "SLOT": "2",
                "IUSE": "B",
                "BDEPEND": "B? ( || ( dev-libs/A:2[B(+)] <dev-libs/A-2[B(+)] <dev-libs/A-2[B(+)] ) )",
            },
            "dev-libs/A-3": {
                "EAPI": "8",
                "SLOT": "3",
                "IUSE": "B",
                "BDEPEND": "B? ( || ( dev-libs/A:3[B(+)] <dev-libs/A-3[B(+)] <dev-libs/A-2[B(+)] ) )",
            },
            "dev-libs/A-4": {
                "EAPI": "8",
                "SLOT": "4",
                "IUSE": "B",
                "BDEPEND": "B? ( || ( dev-libs/A:4[B(+)] <dev-libs/A-4[B(+)] <dev-libs/A-2[B(+)] ) )",
            },
        }

        installed = {
            "dev-libs/A-4": {
                "SLOT": "4",
                "IUSE": "B",
                "USE": "",
                "BDEPEND": "B? ( || ( dev-libs/A:4[B(+)] <dev-libs/A-4[B(+)] <dev-libs/A-2[B(+)] ) )",
            },
        }

        user_config = {
            "package.use": ("dev-libs/A B",),
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A:4"],
                success=True,
                mergelist=["dev-libs/A-1", "dev-libs/A-4"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, user_config=user_config, debug=True
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    @pytest.mark.xfail(reason="bug #947587")
    def testBootstrapChainWithShortcut(self):
        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "8",
                "SLOT": "1",
                "IUSE": "B",
                "BDEPEND": "B? ( || ( <dev-libs/B-2 dev-libs/A:1[B] ) )",
            },
            "dev-libs/A-2": {
                "EAPI": "8",
                "SLOT": "2",
                "IUSE": "B",
                "BDEPEND": "B? ( || ( <dev-libs/B-3 dev-libs/A:2[B] <dev-libs/A-2[B] ) )",
            },
            "dev-libs/A-3": {
                "EAPI": "8",
                "SLOT": "3",
                "IUSE": "B",
                "BDEPEND": "B? ( || ( <dev-libs/B-4 dev-libs/A:3[B] <dev-libs/A-3[B] ) )",
            },
            "dev-libs/A-4": {
                "EAPI": "8",
                "SLOT": "4",
                "IUSE": "B",
                "BDEPEND": "B? ( || ( <dev-libs/B-5 dev-libs/A:4[B] <dev-libs/A-4[B] ) )",
            },
            "dev-libs/B-1": {},
            "dev-libs/B-2": {},
            "dev-libs/B-3": {},
            "dev-libs/B-4": {},
        }

        installed = {
            "dev-libs/A-4": {
                "SLOT": "4",
                "IUSE": "B",
                "USE": "",
                "BDEPEND": "B? ( || ( <dev-libs/B-5 <dev-libs/A:4[B] dev-libs/A-4[B]) )",
            },
        }

        user_config = {
            "package.use": ("dev-libs/A B",),
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A:4"],
                success=True,
                mergelist=["dev-libs/B-4", "dev-libs/A-4"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, user_config=user_config, debug=True
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testBootstrapChainDisruptedShortcutWithShortcutFirst(self):
        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "8",
                "SLOT": "1",
                "IUSE": "B",
                "BDEPEND": "B? ( || ( <dev-libs/B-2 dev-libs/A:1[B] ) )",
            },
            "dev-libs/A-2": {
                "EAPI": "8",
                "SLOT": "2",
                "IUSE": "B C ",
                "BDEPEND": "B? ( || ( <dev-libs/B-3 dev-libs/A:2[B] <dev-libs/A-2[B] ) ) C? ( || ( dev-libs/A:2[C(+)] <dev-libs/A-2[C(+)] <dev-libs/A-2[C(+)] ) )",
            },
            "dev-libs/A-3": {
                "EAPI": "8",
                "SLOT": "3",
                "IUSE": "B C",
                "BDEPEND": "B? ( || ( <dev-libs/B-4 dev-libs/A:3[B] <dev-libs/A-3[B] ) ) C? ( || ( dev-libs/A:3[C(+)] <dev-libs/A-3[C(+)] <dev-libs/A-2[C(+)] ) )",
            },
            "dev-libs/A-4": {
                "EAPI": "8",
                "SLOT": "4",
                "IUSE": "B C",
                "BDEPEND": "B? ( || ( <dev-libs/B-5 dev-libs/A:4[B] <dev-libs/A-4[B] ) ) C? ( || ( dev-libs/A:4[C(+)] <dev-libs/A-4[C(+)] <dev-libs/A-2[C(+)] ) )",
            },
            "dev-libs/B-1": {},
            "dev-libs/B-2": {},
            "dev-libs/B-3": {},
            "dev-libs/B-4": {},
        }

        installed = {
            "dev-libs/A-4": {
                "SLOT": "4",
                "IUSE": "B C",
                "USE": "C",
                "BDEPEND": "B? ( || ( <dev-libs/B-5 dev-libs/A:4[B] <dev-libs/A-4[B] ) ) C? ( || ( dev-libs/A:4[C(+)] <dev-libs/A-4[C(+)] <dev-libs/A-2[C(+)] ) )",
            },
        }

        user_config = {
            "package.use": ("dev-libs/A B C",),
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A:4", "=dev-libs/B-4"],
                success=True,
                mergelist=["dev-libs/B-4", "dev-libs/A-4"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, user_config=user_config, debug=True
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testBootstrapChainDisruptedShortcutWithShortcutLast(self):
        self.__bootstrapChainDisruptedShortcutWithShortcutLastImpl(
            "B? ( || ( dev-libs/A:4[B] =dev-libs/B-4 ) ) C? ( dev-libs/A:4[C] )",
        )

    @pytest.mark.xfail(reason="bug #951296")
    def testBootstrapChainDisruptedShortcutWithShortcutLastAnyOf(self):
        self.__bootstrapChainDisruptedShortcutWithShortcutLastImpl(
            "B? ( || ( dev-libs/A:4[B] =dev-libs/B-4 ) ) C? ( || ( dev-libs/A:4[C] dev-libs/A:4[C] ) )",
        )

    def __bootstrapChainDisruptedShortcutWithShortcutLastImpl(self, bdepend):
        ebuilds = {
            "dev-libs/A-4": {
                "EAPI": "8",
                "SLOT": "4",
                "IUSE": "B C",
                "BDEPEND": bdepend,
            },
            "dev-libs/B-4": {},
        }

        installed = {
            "dev-libs/A-4": {
                "SLOT": "4",
                "IUSE": "B C",
                "USE": "C",
                "BDEPEND": bdepend,
            },
        }

        user_config = {
            "package.use": ("dev-libs/A B C",),
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/A:4", "=dev-libs/B-4"],
                success=True,
                mergelist=["dev-libs/B-4", "dev-libs/A-4"],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, user_config=user_config, debug=True
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2014-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import datetime
import json
import shlex
import subprocess
import sys
import textwrap

import portage
from portage import os, shutil
from portage import _unicode_decode
from portage.const import PORTAGE_PYM_PATH, REPO_REVISIONS, TIMESTAMP_FORMAT
from portage.process import find_binary
from portage.sync.revision_history import get_repo_revision_history
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs


class SyncLocalTestCase(TestCase):
    """
    Test sync with rsync and git, using file:// sync-uri.
    """

    def _must_skip(self):
        if find_binary("rsync") is None:
            self.skipTest("rsync: command not found")
        if find_binary("git") is None:
            self.skipTest("git: command not found")

    def testSyncLocal(self):
        debug = False
        self._must_skip()

        repos_conf = textwrap.dedent(
            """
			[DEFAULT]
			%(default_keys)s
			[test_repo]
			location = %(EPREFIX)s/var/repositories/test_repo
			sync-type = %(sync-type)s
			sync-depth = %(sync-depth)s
			sync-uri = file://%(EPREFIX)s/var/repositories/test_repo_sync
			sync-rcu = %(sync-rcu)s
			sync-rcu-store-dir = %(EPREFIX)s/var/repositories/test_repo_rcu_storedir
			auto-sync = %(auto-sync)s
			volatile = no
			%(repo_extra_keys)s
		"""
        )

        profile = {"eapi": ("5",), "package.use.stable.mask": ("dev-libs/A flag",)}

        ebuilds = {
            "dev-libs/A-0": {"EAPI": "8"},
            "sys-apps/portage-3.0": {"IUSE": "+python_targets_python3_8"},
        }

        installed = {
            "sys-apps/portage-2.3.99": {
                "EAPI": "7",
                "IUSE": "+python_targets_python3_8",
                "USE": "python_targets_python3_8",
            },
        }

        user_config = {"make.conf": ('FEATURES="metadata-transfer"',)}

        playground = ResolverPlayground(
            ebuilds=ebuilds,
            installed=installed,
            profile=profile,
            user_config=user_config,
            debug=debug,
        )
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        homedir = os.path.join(eroot, "home")
        distdir = os.path.join(eprefix, "distdir")
        repo = settings.repositories["test_repo"]
        metadata_dir = os.path.join(repo.location, "metadata")
        rcu_store_dir = os.path.join(eprefix, "var/repositories/test_repo_rcu_storedir")

        cmds = {}
        for cmd in ("egencache", "emerge", "emaint"):
            for bindir in (self.bindir, self.sbindir):
                path = os.path.join(str(bindir), cmd)
                if os.path.exists(path):
                    cmds[cmd] = (portage._python_interpreter, "-b", "-Wd", path)
                    break
            else:
                raise AssertionError(
                    f"{cmd} binary not found in {self.bindir} or {self.sbindir}"
                )

        git_binary = find_binary("git")
        git_cmd = (git_binary,)

        hg_binary = find_binary("hg")
        hg_cmd = (hg_binary,)

        committer_name = "Gentoo Dev"
        committer_email = "gentoo-dev@gentoo.org"

        def repos_set_conf(
            sync_type,
            dflt_keys=None,
            xtra_keys=None,
            auto_sync="yes",
            sync_rcu=False,
            sync_depth=None,
        ):
            env["PORTAGE_REPOSITORIES"] = repos_conf % {
                "EPREFIX": eprefix,
                "sync-type": sync_type,
                "sync-depth": 0 if sync_depth is None else sync_depth,
                "sync-rcu": "yes" if sync_rcu else "no",
                "auto-sync": auto_sync,
                "default_keys": "" if dflt_keys is None else dflt_keys,
                "repo_extra_keys": "" if xtra_keys is None else xtra_keys,
            }

        def alter_ebuild():
            with open(
                os.path.join(repo.location + "_sync", "dev-libs", "A", "A-0.ebuild"),
                "a",
            ) as f:
                f.write("\n")
            bump_timestamp()

        def bump_timestamp():
            bump_timestamp.timestamp += datetime.timedelta(seconds=1)
            with open(
                os.path.join(repo.location + "_sync", "metadata", "timestamp.chk"), "w"
            ) as f:
                f.write(
                    bump_timestamp.timestamp.strftime(
                        f"{TIMESTAMP_FORMAT}\n",
                    )
                )

        bump_timestamp.timestamp = datetime.datetime.now(datetime.timezone.utc)

        bump_timestamp_cmds = ((homedir, bump_timestamp),)

        sync_cmds = (
            (homedir, cmds["emerge"] + ("--sync",)),
            (
                homedir,
                lambda: self.assertTrue(
                    os.path.exists(os.path.join(repo.location, "dev-libs", "A")),
                    "dev-libs/A expected, but missing",
                ),
            ),
            (homedir, cmds["emaint"] + ("sync", "-A")),
        )

        sync_cmds_auto_sync = (
            (homedir, lambda: repos_set_conf("rsync", auto_sync="no")),
            (homedir, cmds["emerge"] + ("--sync",)),
            (
                homedir,
                lambda: self.assertFalse(
                    os.path.exists(os.path.join(repo.location, "dev-libs", "A")),
                    "dev-libs/A found, expected missing",
                ),
            ),
            (homedir, lambda: repos_set_conf("rsync", auto_sync="yes")),
        )

        rename_repo = (
            (homedir, lambda: os.rename(repo.location, repo.location + "_sync")),
        )

        rsync_opts_repos = (
            (homedir, alter_ebuild),
            (
                homedir,
                lambda: repos_set_conf(
                    "rsync",
                    None,
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % shlex.quote(repo.location + "_back"),
                ),
            ),
            (homedir, cmds["emerge"] + ("--sync",)),
            (homedir, lambda: self.assertTrue(os.path.exists(repo.location + "_back"))),
            (homedir, lambda: shutil.rmtree(repo.location + "_back")),
            (homedir, lambda: repos_set_conf("rsync")),
        )

        rsync_opts_repos_default = (
            (homedir, alter_ebuild),
            (
                homedir,
                lambda: repos_set_conf(
                    "rsync",
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % shlex.quote(repo.location + "_back"),
                ),
            ),
            (homedir, cmds["emerge"] + ("--sync",)),
            (homedir, lambda: self.assertTrue(os.path.exists(repo.location + "_back"))),
            (homedir, lambda: shutil.rmtree(repo.location + "_back")),
            (homedir, lambda: repos_set_conf("rsync")),
        )

        rsync_opts_repos_default_ovr = (
            (homedir, alter_ebuild),
            (
                homedir,
                lambda: repos_set_conf(
                    "rsync",
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % shlex.quote(repo.location + "_back_nowhere"),
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % shlex.quote(repo.location + "_back"),
                ),
            ),
            (homedir, cmds["emerge"] + ("--sync",)),
            (homedir, lambda: self.assertTrue(os.path.exists(repo.location + "_back"))),
            (homedir, lambda: shutil.rmtree(repo.location + "_back")),
            (homedir, lambda: repos_set_conf("rsync")),
        )

        rsync_opts_repos_default_cancel = (
            (homedir, alter_ebuild),
            (
                homedir,
                lambda: repos_set_conf(
                    "rsync",
                    "sync-rsync-extra-opts = --backup --backup-dir=%s"
                    % shlex.quote(repo.location + "_back_nowhere"),
                    "sync-rsync-extra-opts = ",
                ),
            ),
            (homedir, cmds["emerge"] + ("--sync",)),
            (
                homedir,
                lambda: self.assertFalse(os.path.exists(repo.location + "_back")),
            ),
            (homedir, lambda: repos_set_conf("rsync")),
        )

        delete_repo_location = (
            (homedir, lambda: shutil.rmtree(repo.user_location)),
            (homedir, lambda: os.mkdir(repo.user_location)),
        )

        delete_rcu_store_dir = ((homedir, lambda: shutil.rmtree(rcu_store_dir)),)

        revert_rcu_layout = (
            (
                homedir,
                lambda: os.rename(repo.user_location, repo.user_location + ".bak"),
            ),
            (
                homedir,
                lambda: os.rename(
                    os.path.realpath(repo.user_location + ".bak"), repo.user_location
                ),
            ),
            (homedir, lambda: os.unlink(repo.user_location + ".bak")),
            (homedir, lambda: shutil.rmtree(repo.user_location + "_rcu_storedir")),
        )

        upstream_git_commit = (
            (
                repo.location + "_sync",
                git_cmd + ("commit", "--allow-empty", "-m", "test empty commit"),
            ),
            (
                repo.location + "_sync",
                git_cmd + ("commit", "--allow-empty", "-m", "test empty commit 2"),
            ),
        )

        delete_sync_repo = ((homedir, lambda: shutil.rmtree(repo.location + "_sync")),)

        git_repo_create = (
            (
                repo.location,
                git_cmd
                + (
                    "config",
                    "--global",
                    "user.name",
                    committer_name,
                ),
            ),
            (
                repo.location,
                git_cmd
                + (
                    "config",
                    "--global",
                    "user.email",
                    committer_email,
                ),
            ),
            (repo.location, git_cmd + ("init-db",)),
            # Ensure manifests and cache are valid after
            # previous calls to alter_ebuild.
            (
                homedir,
                cmds["egencache"]
                + (
                    f"--repo={repo.name}",
                    "--update",
                    "--update-manifests",
                    "--sign-manifests=n",
                    "--strict-manifests=n",
                    f"--repositories-configuration={settings['PORTAGE_REPOSITORIES']}",
                    f"--jobs={portage.util.cpuinfo.get_cpu_count()}",
                ),
            ),
            (repo.location, git_cmd + ("add", ".")),
            (repo.location, git_cmd + ("commit", "-a", "-m", "add whole repo")),
        )

        sync_type_git = ((homedir, lambda: repos_set_conf("git")),)

        sync_type_git_shallow = (
            (homedir, lambda: repos_set_conf("git", sync_depth=1)),
        )

        sync_rsync_rcu = ((homedir, lambda: repos_set_conf("rsync", sync_rcu=True)),)

        delete_git_dir = (
            (homedir, lambda: shutil.rmtree(os.path.join(repo.location, ".git"))),
        )

        def get_revision_history(sync_type="git"):
            # Override volatile to False here because it gets set
            # True by RepoConfig when repo.location is not root
            # or portage owned.
            try:
                volatile_orig = repo.volatile
                repo.volatile = False
                sync_type_orig = repo.sync_type
                repo.sync_type = sync_type
                revision_history = get_repo_revision_history(eroot, repos=[repo])
            finally:
                repo.sync_type = sync_type_orig
                repo.volatile = volatile_orig

            return revision_history

        repo_revisions_cmds = (
            (homedir, lambda: self.assertTrue(bool(get_revision_history()))),
            (
                homedir,
                lambda: self.assertTrue(
                    os.path.exists(os.path.join(eroot, REPO_REVISIONS))
                ),
            ),
            (homedir, cmds["emaint"] + ("revisions", f"--purgerepos={repo.name}")),
            (
                homedir,
                lambda: self.assertFalse(
                    os.path.exists(os.path.join(eroot, REPO_REVISIONS))
                ),
            ),
            (homedir, lambda: self.assertTrue(bool(get_revision_history()))),
            (
                homedir,
                lambda: self.assertTrue(
                    os.path.exists(os.path.join(eroot, REPO_REVISIONS))
                ),
            ),
            (homedir, cmds["emaint"] + ("revisions", "--purgeallrepos")),
            (
                homedir,
                lambda: self.assertFalse(
                    os.path.exists(os.path.join(eroot, REPO_REVISIONS))
                ),
            ),
            (homedir, lambda: self.assertTrue(bool(get_revision_history()))),
        )

        def assert_latest_rev_in_packages_index(positive):
            """
            If we build a binary package then its REPO_REVISIONS should
            propagate into $PKGDIR/Packages as long as it results in
            forward progress according to the repo revision history.
            """
            revision_history = get_revision_history()
            prefix = "REPO_REVISIONS:"
            header_repo_revisions = None
            try:
                with open(
                    os.path.join(settings["PKGDIR"], "Packages"), encoding="utf8"
                ) as f:
                    for line in f:
                        if line.startswith(prefix):
                            header_repo_revisions = line[len(prefix) :].strip()
                            break
            except FileNotFoundError:
                pass

            if positive:
                self.assertFalse(header_repo_revisions is None)

            if header_repo_revisions is None:
                header_repo_revisions = {}
            else:
                header_repo_revisions = json.loads(header_repo_revisions)

            (self.assertEqual if positive else self.assertNotEqual)(
                revision_history.get(repo.name, [False])[0],
                header_repo_revisions.get(repo.name, None),
            )

        pkgindex_revisions_cmds = (
            (homedir, lambda: assert_latest_rev_in_packages_index(False)),
            (homedir, cmds["emerge"] + ("-B", "dev-libs/A")),
            (homedir, lambda: assert_latest_rev_in_packages_index(True)),
        )

        def hg_init_global_config():
            with open(os.path.join(homedir, ".hgrc"), "w") as f:
                f.write(f"[ui]\nusername = {committer_name} <{committer_email}>\n")

        hg_repo_create = (
            (repo.location, hg_init_global_config),
            (repo.location, hg_cmd + ("init",)),
            (repo.location, hg_cmd + ("add", ".")),
            (repo.location, hg_cmd + ("commit", "-A", "-m", "add whole repo")),
        )

        sync_type_mercurial = ((homedir, lambda: repos_set_conf("mercurial")),)

        def append_newline(path):
            with open(path, "a") as f:
                f.write("\n")

        upstream_hg_commit = (
            (
                repo.location + "_sync",
                lambda: append_newline(
                    os.path.join(repo.location + "_sync", "metadata/layout.conf")
                ),
            ),
            (
                repo.location + "_sync",
                hg_cmd + ("commit", "metadata/layout.conf", "-m", "test empty commit"),
            ),
            (
                repo.location + "_sync",
                lambda: append_newline(
                    os.path.join(repo.location + "_sync", "metadata/layout.conf")
                ),
            ),
            (
                repo.location + "_sync",
                hg_cmd
                + ("commit", "metadata/layout.conf", "-m", "test empty commit 2"),
            ),
        )

        if hg_binary is None:
            mercurial_tests = ()
        else:
            mercurial_tests = (
                delete_sync_repo
                + delete_git_dir
                + hg_repo_create
                + sync_type_mercurial
                + rename_repo
                + sync_cmds
                + upstream_hg_commit
                + sync_cmds
            )

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = settings.environ()
        env.update(
            {
                "PORTAGE_OVERRIDE_EPREFIX": eprefix,
                "DISTDIR": distdir,
                "GENTOO_COMMITTER_NAME": committer_name,
                "GENTOO_COMMITTER_EMAIL": committer_email,
                "HOME": homedir,
                "PORTAGE_INST_GID": str(os.getgid()),
                "PORTAGE_INST_UID": str(os.getuid()),
                "PORTAGE_GRPNAME": os.environ["PORTAGE_GRPNAME"],
                "PORTAGE_USERNAME": os.environ["PORTAGE_USERNAME"],
                "PYTHONDONTWRITEBYTECODE": os.environ.get(
                    "PYTHONDONTWRITEBYTECODE", ""
                ),
                "PYTHONPATH": pythonpath,
            }
        )

        repos_set_conf("rsync")

        if os.environ.get("SANDBOX_ON") == "1":
            # avoid problems from nested sandbox instances
            env["FEATURES"] = "-sandbox -usersandbox"

        dirs = [homedir, metadata_dir]
        try:
            for d in dirs:
                ensure_dirs(d)

            timestamp_path = os.path.join(metadata_dir, "timestamp.chk")
            with open(timestamp_path, "w") as f:
                f.write(
                    bump_timestamp.timestamp.strftime(
                        f"{TIMESTAMP_FORMAT}\n",
                    )
                )

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for cwd, cmd in (
                rename_repo
                + sync_cmds_auto_sync
                + sync_cmds
                + rsync_opts_repos
                + rsync_opts_repos_default
                + rsync_opts_repos_default_ovr
                + rsync_opts_repos_default_cancel
                + bump_timestamp_cmds
                + sync_rsync_rcu
                + sync_cmds
                + delete_rcu_store_dir
                + sync_cmds
                + revert_rcu_layout
                + delete_repo_location
                + sync_cmds
                + sync_cmds
                + bump_timestamp_cmds
                + sync_cmds
                + revert_rcu_layout
                + delete_sync_repo
                + git_repo_create
                + sync_type_git
                + rename_repo
                + sync_cmds
                + upstream_git_commit
                + sync_cmds
                + sync_type_git_shallow
                + upstream_git_commit
                + sync_cmds
                + repo_revisions_cmds
                + pkgindex_revisions_cmds
                + mercurial_tests
            ):
                if hasattr(cmd, "__call__"):
                    cmd()
                    continue

                abs_cwd = os.path.join(repo.location, cwd)
                proc = subprocess.Popen(cmd, cwd=abs_cwd, env=env, stdout=stdout)

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != os.EX_OK:
                        for line in output:
                            sys.stderr.write(_unicode_decode(line))

                self.assertEqual(
                    os.EX_OK,
                    proc.returncode,
                    f"{cmd} failed in {cwd}",
                )

        finally:
            playground.cleanup()



# validators.py Portage File Loader Code
# Copyright 2007-2020 Gentoo Authors

from portage.dep import isvalidatom

ValidAtomValidator = isvalidatom


def PackagesFileValidator(atom):
    """This function mutates atoms that begin with - or *
    It then checks to see if that atom is valid, and if
    so returns True, else it returns False.

    Args:
        atom: a string representing an atom such as sys-apps/portage-2.1
    """
    if atom.startswith("*") or atom.startswith("-"):
        atom = atom[1:]
    if not isvalidatom(atom):
        return False
    return True



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class SlotOperatorUpdateProbeParentDowngradeTestCase(TestCase):
    def testSlotOperatorUpdateProbeParentDowngrade(self):
        ebuilds = {
            "net-nds/openldap-2.4.40-r3": {
                "EAPI": "5",
                "RDEPEND": "<sys-libs/db-6.0:= "
                + "|| ( sys-libs/db:5.3 sys-libs/db:5.1 )",
            },
            "net-nds/openldap-2.4.40": {"EAPI": "5", "RDEPEND": "sys-libs/db"},
            "sys-libs/db-6.0": {
                "SLOT": "6.0",
            },
            "sys-libs/db-5.3": {
                "SLOT": "5.3",
            },
        }

        installed = {
            "net-nds/openldap-2.4.40-r3": {
                "EAPI": "5",
                "RDEPEND": "<sys-libs/db-6.0:5.3/5.3= "
                + "|| ( sys-libs/db:5.3 sys-libs/db:5.1 )",
            },
            "sys-libs/db-6.0": {
                "SLOT": "6.0",
            },
            "sys-libs/db-5.3": {
                "SLOT": "5.3",
            },
        }

        world = ("net-nds/openldap",)

        test_cases = (
            # bug 528610 - openldap rebuild was triggered
            # inappropriately, due to slot_operator_update_probe
            # selecting an inappropriate replacement parent of
            # a lower version than desired.
            ResolverPlaygroundTestCase(
                ["@world"],
                success=True,
                options={"--update": True, "--deep": True},
                mergelist=[],
            ),
        )

        playground = ResolverPlayground(
            ebuilds=ebuilds, installed=installed, world=world, debug=False
        )
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# test_stackDictList.py -- Portage Unit Testing Functionality
# Copyright 2006 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase


class StackDictListTestCase(TestCase):
    def testStackDictList(self):
        from portage.util import stack_dictlist

        tests = [
            ({"a": "b"}, {"x": "y"}, False, {"a": ["b"], "x": ["y"]}),
            ({"KEYWORDS": ["alpha", "x86"]}, {"KEYWORDS": ["-*"]}, True, {}),
            (
                {"KEYWORDS": ["alpha", "x86"]},
                {"KEYWORDS": ["-x86"]},
                True,
                {"KEYWORDS": ["alpha"]},
            ),
        ]
        for test in tests:
            self.assertEqual(
                stack_dictlist([test[0], test[1]], incremental=test[2]), test[3]
            )



# Copyright 1999-2024 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

# Don't use the unicode-wrapped os and shutil modules here since
# the whole _selinux module itself will be wrapped.
import os
import shutil
import warnings
from functools import partial

try:
    import selinux
except ImportError:
    selinux = None

import portage
from portage import _encodings
from portage import _native_string, _unicode_encode
from portage.localization import _


def copyfile(src, dest):
    src = _native_string(src, encoding=_encodings["fs"], errors="strict")
    dest = _native_string(dest, encoding=_encodings["fs"], errors="strict")
    (rc, ctx) = selinux.lgetfilecon(src)
    if rc < 0:
        raise OSError(_('copyfile: Failed getting context of "%s".') % src)

    setfscreate(ctx)
    try:
        shutil.copyfile(src, dest)
    finally:
        setfscreate()


def getcontext():
    (rc, ctx) = selinux.getcon()
    if rc < 0:
        raise OSError(_("getcontext: Failed getting current process context."))

    return ctx


def is_selinux_enabled():
    return selinux.is_selinux_enabled()


def mkdir(target, refdir):
    refdir = _native_string(refdir, encoding=_encodings["fs"], errors="strict")
    (rc, ctx) = selinux.getfilecon(refdir)
    if rc < 0:
        raise OSError(
            _('mkdir: Failed getting context of reference directory "%s".') % refdir
        )

    setfscreate(ctx)
    try:
        os.mkdir(_unicode_encode(target, encoding=_encodings["fs"], errors="strict"))
    finally:
        setfscreate()


def rename(src, dest):
    src = _native_string(src, encoding=_encodings["fs"], errors="strict")
    (rc, ctx) = selinux.lgetfilecon(src)
    if rc < 0:
        raise OSError(_('rename: Failed getting context of "%s".') % src)

    setfscreate(ctx)
    try:
        os.rename(
            _unicode_encode(src, encoding=_encodings["fs"], errors="strict"),
            _unicode_encode(dest, encoding=_encodings["fs"], errors="strict"),
        )
    finally:
        setfscreate()


def settype(newtype):
    try:
        ret = getcontext().split(":")
        ret[2] = newtype
        return ":".join(ret)
    except IndexError:
        warnings.warn(f"Invalid SELinux context: {getcontext()}")
        return None


def setexec(ctx="\n"):
    ctx = _native_string(ctx, encoding=_encodings["content"], errors="strict")
    rc = 0
    try:
        rc = selinux.setexeccon(ctx)
    except OSError:
        msg = _(
            "Failed to set new SELinux execution context. "
            + "Is your current SELinux context allowed to run Portage?"
        )
        if selinux.security_getenforce() == 1:
            raise OSError(msg)
        else:
            portage.writemsg(f"!!! {msg}\n", noiselevel=-1)

    if rc < 0:
        if selinux.security_getenforce() == 1:
            raise OSError(_('Failed setting exec() context "%s".') % ctx)
        else:
            portage.writemsg(
                "!!! " + _('Failed setting exec() context "%s".') % ctx, noiselevel=-1
            )


def setfscreate(ctx="\n"):
    ctx = _native_string(ctx, encoding=_encodings["content"], errors="strict")
    if selinux.setfscreatecon(ctx) < 0:
        raise OSError(_('setfscreate: Failed setting fs create context "%s".') % ctx)


class spawn_wrapper:
    """
    Create a wrapper function for the given spawn function. When the wrapper
    is called, it will adjust the arguments such that setexec() to be called
    *after* the fork (thereby avoiding any interference with concurrent
    threads in the calling process).
    """

    __slots__ = ("_con", "_spawn_func")

    def __init__(self, spawn_func, selinux_type):
        self._spawn_func = spawn_func
        selinux_type = _native_string(
            selinux_type, encoding=_encodings["content"], errors="strict"
        )
        self._con = settype(selinux_type)

    def __call__(self, *args, **kwargs):
        if self._con is not None:
            pre_exec = partial(setexec, self._con)
            kwargs["pre_exec"] = (
                portage.process._chain_pre_exec_fns(pre_exec, kwargs["pre_exec"])
                if kwargs.get("pre_exec")
                else pre_exec
            )

        return self._spawn_func(*args, **kwargs)


def symlink(target, link, reflnk):
    reflnk = _native_string(reflnk, encoding=_encodings["fs"], errors="strict")
    (rc, ctx) = selinux.lgetfilecon(reflnk)
    if rc < 0:
        raise OSError(
            _('symlink: Failed getting context of reference symlink "%s".') % reflnk
        )

    setfscreate(ctx)
    try:
        os.symlink(
            _unicode_encode(target, encoding=_encodings["fs"], errors="strict"),
            _unicode_encode(link, encoding=_encodings["fs"], errors="strict"),
        )
    finally:
        setfscreate()



# Copyright 2014 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

doc = """SVN plug-in module for portage.
Performs a svn up on repositories."""
__doc__ = doc[:]

from portage.localization import _
from portage.sync.config_checks import CheckSyncConfig
from portage.util import writemsg_level


module_spec = {
    "name": "svn",
    "description": doc,
    "provides": {
        "svn-module": {
            "name": "svn",
            "sourcefile": "svn",
            "class": "SVNSync",
            "description": doc,
            "functions": ["sync", "new", "exists"],
            "func_desc": {
                "sync": "Performs a svn up on the repository",
                "new": "Creates the new repository at the specified location",
                "exists": "Returns a boolean of whether the specified dir "
                + "exists and is a valid SVN repository",
            },
            "validate_config": CheckSyncConfig,
            "module_specific_options": (),
        }
    },
}



# Copyright 2011-2019 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)


class KeywordsTestCase(TestCase):
    def testStableConfig(self):
        # Only accept stable keywords for a particular ARCH.

        user_config = {
            "package.accept_keywords": ("*/* -* x86",),
        }

        ebuilds = {
            "app-misc/A-1": {"KEYWORDS": "x86"},
            "app-misc/B-1": {"KEYWORDS": "~x86"},
            "app-misc/C-1": {"KEYWORDS": "*"},
            "app-misc/D-1": {"KEYWORDS": "~*"},
            "app-misc/E-1": {"KEYWORDS": "arm"},
            "app-misc/F-1": {"KEYWORDS": "~arm"},
            "app-misc/G-1": {"KEYWORDS": ""},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"], success=True, mergelist=["app-misc/A-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/B"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/B-1",),
                mergelist=["app-misc/B-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/C"], success=True, mergelist=["app-misc/C-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/D"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/D-1",),
                mergelist=["app-misc/D-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/E"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/E-1",),
                mergelist=["app-misc/E-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/F"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/F-1",),
                mergelist=["app-misc/F-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/G"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/G-1",),
                mergelist=["app-misc/G-1"],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, user_config=user_config)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testAnyStableConfig(self):
        # Accept stable keywords for any ARCH.

        user_config = {
            "package.accept_keywords": ("*/* -* *",),
        }

        ebuilds = {
            "app-misc/A-1": {"KEYWORDS": "x86"},
            "app-misc/B-1": {"KEYWORDS": "~x86"},
            "app-misc/C-1": {"KEYWORDS": "*"},
            "app-misc/D-1": {"KEYWORDS": "~*"},
            "app-misc/E-1": {"KEYWORDS": "arm"},
            "app-misc/F-1": {"KEYWORDS": "~arm"},
            "app-misc/G-1": {"KEYWORDS": ""},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"], success=True, mergelist=["app-misc/A-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/B"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/B-1",),
                mergelist=["app-misc/B-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/C"], success=True, mergelist=["app-misc/C-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/D"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/D-1",),
                mergelist=["app-misc/D-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/E"], success=True, mergelist=["app-misc/E-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/F"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/F-1",),
                mergelist=["app-misc/F-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/G"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/G-1",),
                mergelist=["app-misc/G-1"],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, user_config=user_config)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testUnstableConfig(self):
        # Accept stable and unstable keywords for a particular ARCH.

        user_config = {
            "package.accept_keywords": ("*/* -* x86 ~x86",),
        }

        ebuilds = {
            "app-misc/A-1": {"KEYWORDS": "x86"},
            "app-misc/B-1": {"KEYWORDS": "~x86"},
            "app-misc/C-1": {"KEYWORDS": "*"},
            "app-misc/D-1": {"KEYWORDS": "~*"},
            "app-misc/E-1": {"KEYWORDS": "arm"},
            "app-misc/F-1": {"KEYWORDS": "~arm"},
            "app-misc/G-1": {"KEYWORDS": ""},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"], success=True, mergelist=["app-misc/A-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/B"], success=True, mergelist=["app-misc/B-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/C"], success=True, mergelist=["app-misc/C-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/D"], success=True, mergelist=["app-misc/D-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/E"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/E-1",),
                mergelist=["app-misc/E-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/F"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/F-1",),
                mergelist=["app-misc/F-1"],
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/G"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/G-1",),
                mergelist=["app-misc/G-1"],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, user_config=user_config)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testAnyUnstableConfig(self):
        # Accept unstable keywords for any ARCH.

        user_config = {
            "package.accept_keywords": ("*/* -* * ~*",),
        }

        ebuilds = {
            "app-misc/A-1": {"KEYWORDS": "x86"},
            "app-misc/B-1": {"KEYWORDS": "~x86"},
            "app-misc/C-1": {"KEYWORDS": "*"},
            "app-misc/D-1": {"KEYWORDS": "~*"},
            "app-misc/E-1": {"KEYWORDS": "arm"},
            "app-misc/F-1": {"KEYWORDS": "~arm"},
            "app-misc/G-1": {"KEYWORDS": ""},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"], success=True, mergelist=["app-misc/A-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/B"], success=True, mergelist=["app-misc/B-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/C"], success=True, mergelist=["app-misc/C-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/D"], success=True, mergelist=["app-misc/D-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/E"], success=True, mergelist=["app-misc/E-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/F"], success=True, mergelist=["app-misc/F-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/G"],
                success=False,
                options={"--autounmask": True},
                unstable_keywords=("app-misc/G-1",),
                mergelist=["app-misc/G-1"],
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, user_config=user_config)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()

    def testIgnoreKeywordsConfig(self):
        # Ignore keywords entirely (accept **)

        user_config = {
            "package.accept_keywords": ("*/* -* **",),
        }

        ebuilds = {
            "app-misc/A-1": {"KEYWORDS": "x86"},
            "app-misc/B-1": {"KEYWORDS": "~x86"},
            "app-misc/C-1": {"KEYWORDS": "*"},
            "app-misc/D-1": {"KEYWORDS": "~*"},
            "app-misc/E-1": {"KEYWORDS": "arm"},
            "app-misc/F-1": {"KEYWORDS": "~arm"},
            "app-misc/G-1": {"KEYWORDS": ""},
        }

        test_cases = (
            ResolverPlaygroundTestCase(
                ["app-misc/A"], success=True, mergelist=["app-misc/A-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/B"], success=True, mergelist=["app-misc/B-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/C"], success=True, mergelist=["app-misc/C-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/D"], success=True, mergelist=["app-misc/D-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/E"], success=True, mergelist=["app-misc/E-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/F"], success=True, mergelist=["app-misc/F-1"]
            ),
            ResolverPlaygroundTestCase(
                ["app-misc/G"], success=True, mergelist=["app-misc/G-1"]
            ),
        )

        playground = ResolverPlayground(ebuilds=ebuilds, user_config=user_config)
        try:
            for test_case in test_cases:
                playground.run_TestCase(test_case)
                self.assertEqual(test_case.test_success, True, test_case.fail_msg)
        finally:
            playground.cleanup()



# Copyright 2011-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from itertools import chain

from portage.const import PORTAGE_PYM_PATH, PORTAGE_PYM_PACKAGES
from portage.tests import TestCase
from portage import os
from portage import _encodings
from portage import _unicode_decode


class ImportModulesTestCase(TestCase):
    def testImportModules(self):
        expected_failures = frozenset(())

        iters = (
            self._iter_modules(os.path.join(PORTAGE_PYM_PATH, x))
            for x in PORTAGE_PYM_PACKAGES
        )
        for mod in chain(*iters):
            try:
                __import__(mod)
            except ImportError as e:
                if mod not in expected_failures:
                    self.assertTrue(False, f"failed to import '{mod}': {e}")
                del e

    def _iter_modules(self, base_dir):
        for parent, dirs, files in os.walk(base_dir):
            parent = _unicode_decode(parent, encoding=_encodings["fs"], errors="strict")
            parent_mod = parent[len(PORTAGE_PYM_PATH) + 1 :]
            parent_mod = parent_mod.replace("/", ".")
            for x in files:
                x = _unicode_decode(x, encoding=_encodings["fs"], errors="strict")
                if x[-3:] != ".py":
                    continue
                x = x[:-3]
                if x[-8:] == "__init__":
                    x = parent_mod
                else:
                    x = parent_mod + "." + x
                yield x
